"""
Univariate Sine Wave
-----------------
"""

import numpy as np
from oats.generator._base import Generator


def _series_segmentation(data, stepsize=1):
    return np.split(data, np.where(np.diff(data) != stepsize)[0] + 1)


def _sine(length, freq=0.04, coef=1.5, offset=0.0, noise_amp=0.05):
    # timestamp = np.linspace(0, 10, length)
    timestamp = np.arange(length)
    value = np.sin(2 * np.pi * freq * timestamp)
    if noise_amp != 0:
        noise = np.random.normal(0, 1, length)
        value = value + noise_amp * noise
    value = coef * value + offset
    return value


def _square_sine(level=5, length=500, freq=0.04, coef=1.5, offset=0.0, noise_amp=0.05):
    value = np.zeros(length)
    for i in range(level):
        value += (
            1
            / (2 * i + 1)
            * _sine(
                length=length,
                freq=freq * (2 * i + 1),
                coef=coef,
                offset=offset,
                noise_amp=noise_amp,
            )
        )
    return value


def _collective_global_synthetic(length, base, coef=1.5, noise_amp=0.005):
    value = []
    norm = np.linalg.norm(base)
    base = base / norm
    num = int(length / len(base))
    for i in range(num):
        value.extend(base)
    residual = length - len(value)
    value.extend(base[:residual])
    value = np.array(value)
    noise = np.random.normal(0, 1, length)
    value = coef * value + noise_amp * noise
    return value


class UnivariateWaveGenerator(Generator):
    """Generates a univariate signal and append anomalies as desired.


    Attributes:
        train: numpy array of generated train data
        test: numpy array of generated test data
        label: numpy array of labels

    Example:
    ```
        generator = UnivariateWaveGenerator(1000)
        generator.point_global_outliers(0.05, 2, 50)
        train, test, label = generator.get_dataset()
    ```
    """

    _BEHAVIOR_CONFIG = {"freq": 0.04, "coef": 1.5, "offset": 0.0, "noise_amp": 0.05}

    def __init__(
        self,
        stream_length: int,
        train_ratio: float = 0.2,
        behavior=_sine,
        behavior_config: dict = _BEHAVIOR_CONFIG,
    ):
        """
        Args:
            stream_length (int): Total length of `train` and `test` combined
            train_ratio (float, optional): Length of train relative to `stream_length`. Defaults to 0.2.
            behavior (_type_, optional): Underlying signal generating function. Defaults to sine.
            behavior_config (dict, optional): Default config for . Defaults to BEHAVIOR_CONFIG.
        """
        self.STREAM_LENGTH = stream_length
        self.train_ratio = train_ratio
        self.split = int(self.STREAM_LENGTH * self.train_ratio)

        self.behavior = behavior
        self.behavior_config = behavior_config if behavior_config is not None else {}

        self.train = None
        self.data = None
        self.label = None

        self.data_origin = None
        self.timestamp = np.arange(self.STREAM_LENGTH)

        self._generate_timeseries()

    def _generate_timeseries(self):
        self.behavior_config["length"] = self.STREAM_LENGTH
        self.data = self.behavior(**self.behavior_config)

        self.train = self.data[: self.split]
        self.test = self.data[self.split :]

        self.TEST_LENGTH = len(self.test)

        self.test_orig = self.test.copy()
        self.label = np.zeros(self.TEST_LENGTH, dtype=int)

    def point_global_outliers(self, ratio, factor, radius):
        """Add point global outliers to original data

        Args:
            ratio: what ratio outliers will be added
            factor: the larger, the outliers are farther from inliers
            radius: the radius of collective outliers range
        """
        position = (
            np.random.rand(round(self.TEST_LENGTH * ratio)) * self.TEST_LENGTH
        ).astype(int)
        maximum, minimum = max(self.test), min(self.test)
        for i in position:
            local_std = self.test_orig[
                max(0, i - radius) : min(i + radius, self.TEST_LENGTH)
            ].std()
            self.test[i] = self.test_orig[i] * factor * local_std
            if 0 <= self.test[i] < maximum:
                self.test[i] = maximum
            if 0 > self.test[i] > minimum:
                self.test[i] = minimum
            self.label[i] = 1

    def point_contextual_outliers(self, ratio, factor, radius):
        """Add point contextual outliers to original data

        Args:
            ratio: what ratio outliers will be added
            factor: the larger, the outliers are farther from inliers
                    Notice: point contextual outliers will not exceed the range of [min, max] of original data
            radius: the radius of collective outliers range
        """
        position = (
            np.random.rand(round(self.TEST_LENGTH * ratio)) * self.TEST_LENGTH
        ).astype(int)
        maximum, minimum = max(self.test), min(self.test)
        for i in position:
            local_std = self.test[
                max(0, i - radius) : min(i + radius, self.TEST_LENGTH)
            ].std()
            self.test[i] = self.test[i] * factor * local_std
            if self.test[i] > maximum:
                self.test[i] = maximum * min(
                    0.95, abs(np.random.normal(0, 0.5))
                )  # previous(0, 1)
            if self.test[i] < minimum:
                self.test[i] = minimum * min(0.95, abs(np.random.normal(0, 0.5)))

            self.label[i] = 1

    def collective_shapelet_outliers(
        self,
        ratio,
        radius,
        option="square",
        coef=3.0,
        noise_amp=0.0,
        level=5,
        freq=0.04,
        offset=0.0,  # only used when option=='square'
        base=[
            0.0,
        ],
    ):  # only used when option=='other'
        """Add collective global outliers to original data

        Args:
            ratio: what ratio outliers will be added
            radius: the radius of collective outliers range
            option: if 'square': 'level' 'freq' and 'offset' are used to generate square sine wave
                    if 'other': 'base' is used to generate outlier shape
            level: how many sine waves will square_wave synthesis
            base: a list of values that we want to substitute inliers when we generate outliers
        """
        position = (
            np.random.rand(round(self.TEST_LENGTH * ratio / (2 * radius)))
            * self.TEST_LENGTH
        ).astype(int)

        valid_option = {"square", "other"}
        if option not in valid_option:
            raise ValueError("'option' must be one of %r." % valid_option)

        if option == "square":
            sub_data = _square_sine(
                level=level,
                length=self.TEST_LENGTH,
                freq=freq,
                coef=coef,
                offset=offset,
                noise_amp=noise_amp,
            )
        else:
            sub_data = _collective_global_synthetic(
                length=self.TEST_LENGTH, base=base, coef=coef, noise_amp=noise_amp
            )
        for i in position:
            start, end = max(0, i - radius), min(self.TEST_LENGTH, i + radius)
            self.test[start:end] = sub_data[start:end]
            self.label[start:end] = 1

    def collective_trend_outliers(self, ratio, factor, radius):
        """Add collective trend outliers to original data

        Args:
            ratio: what ratio outliers will be added
            factor: how dramatic will the trend be
            radius: the radius of collective outliers range
        """
        position = (
            np.random.rand(round(self.TEST_LENGTH * ratio / (2 * radius)))
            * self.TEST_LENGTH
        ).astype(int)
        for i in position:
            start, end = max(0, i - radius), min(self.TEST_LENGTH, i + radius)
            slope = np.random.choice([-1, 1]) * factor * np.arange(end - start)
            self.test[start:end] = self.test_orig[start:end] + slope
            self.test[end:] = self.test[end:] + slope[-1]
            self.label[start:end] = 1

    def collective_seasonal_outliers(self, ratio, factor, radius):
        """Add collective seasonal outliers to original data

        Args:
            ratio: what ratio outliers will be added
            factor: how many times will frequency multiple
            radius: the radius of collective outliers range
        """
        position = (
            np.random.rand(round(self.TEST_LENGTH * ratio / (2 * radius)))
            * self.TEST_LENGTH
        ).astype(int)
        seasonal_config = self.behavior_config
        seasonal_config["freq"] = factor * self.behavior_config["freq"]
        for i in position:
            start, end = max(0, i - radius), min(self.TEST_LENGTH, i + radius)
            self.test[start:end] = self.behavior(**seasonal_config)[start:end]
            self.label[start:end] = 1

    def get_dataset(self):
        """
        Returns:
            train: numpy array of train data
            test: numpy array of test data
            label: numpy array of labels
        """
        return self.train, self.test, self.label



from typing import Tuple
import numpy.typing as npt


class Generator():
    """Base class for Generators
    Because of complex behavior of generators. Method signatures are not strictly enforced.
    But must implement `get_dataset()` method that returns `(train, test, label)`
    """

    def get_dataset(
        self, *args, **kwargs
    ) -> Tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
        raise NotImplementedError



"""
Temporal Fusion Transformer (TFT)
-----------------
"""
from typing import Any
from functools import partial

from darts import models
import numpy as np
import numpy.typing as npt
import optuna

from oats.models._darts_model import DartsModel


class TFTModel(DartsModel):
    """TFT Model (Temporal Fusion Transformer)

    Using TFT as a predictor. Anomalies scores are deviations from predictions.

    Reference: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.tft_model.html
    """

    def __init__(
        self,
        window: int = 10,
        n_steps: int = 1,
        use_gpu: bool = 1,
        val_split: float = 0.2,
        **kwargs
    ):
        """
        initialization also accepts any parameters used by: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.tft_model.html

        Args:
            window (int, optional): rolling window size to feed into the predictor. Defaults to 10.
            n_steps (int, optional): number of steps to predict forward. Defaults to 1.
            use_gpu (bool, optional): whether to use GPU. Defaults to False.
            val_split (float, optional): proportion of data points reserved for validation. Defaults to 0.2.
        """

        model = models.TFTModel

        super().__init__(model, window, n_steps, use_gpu, val_split, **kwargs)

    def _model_objective(self, trial, train_data: npt.NDArray[Any]):
        params = {
            "add_relative_index": trial.suggest_categorical(
                "add_relative_idex", [True]
            ),
            "hidden_size": trial.suggest_int("hidden_size", 8, 128),
            "lstm_layers": trial.suggest_int("lstm_layers", 1, 32),
            "num_attention_heads": trial.suggest_int("num_attention_heads", 2, 8),
            "hidden_continuous_size": trial.suggest_int(
                "hidden_continuous_size", 4, 32
            ),
            "full_attention": trial.suggest_categorical(
                "full_attention", [True, False]
            ),
            "dropout": trial.suggest_float("dropout", 0.0, 0.3),
        }

        return self._get_hyperopt_res(params, train_data)



"""
Jenks Natural Breaks
-----------------
"""

import numpy as np
import jenkspy

from oats.threshold._base import Threshold


class JenksThreshold(Threshold):
    """Getting threshold using 1-D clustering (Jenks Natural Breaks).
    No need to call `fit()`.
    """

    def fit(self, *args, **kwargs):
        return

    def get_threshold(self, data, n_partitions=20):
        """
        Args:
            data (np.ndarray): numpy array of scores
            n_partitions (int, optional): number of partitions to fit. Defaults to 20.

        Returns:
            np.ndarray: thresholds
        """
        multivar = True if data.ndim > 1 and data.shape[1] > 1 else False
        if multivar:
            return self._handle_multivariate(
                data, [self] * data.shape[1], n_partitions=n_partitions
            )
        thres = jenkspy.jenks_breaks(data.flatten(), nb_class=n_partitions)[-2]
        return np.tile(thres, len(data))



"""
Models
-----------------
"""

import tensorflow as tf

gpu_devices = tf.config.experimental.list_physical_devices("GPU")
for device in gpu_devices:
    tf.config.experimental.set_memory_growth(device, True)

# Predictive Models
from oats.models.predictive.lightgbm import LightGBMModel
from oats.models.predictive.nbeats import NBEATSModel
from oats.models.predictive.nhits import NHiTSModel
from oats.models.predictive.randomforest import RandomForestModel
from oats.models.predictive.rnn import RNNModel
from oats.models.predictive.tcn import TCNModel
from oats.models.predictive.tft import TFTModel
from oats.models.predictive.transformer import TransformerModel
from oats.models.predictive.ma import MovingAverageModel
from oats.models.predictive.arima import ARIMAModel
from oats.models.predictive.regression import RegressionModel
from oats.models.predictive.fluxev import FluxEVModel

# Reconstruction
from oats.models.reconstruction.tranad import TranADModel
from oats.models.reconstruction.vae import VAEModel

# Distance-Based
from oats.models.distance.iforest import IsolationForestModel
from oats.models.distance.matrixprofile import MatrixProfileModel

# Rule-Based
from oats.models.rule_based.quantile import QuantileModel



"""
ARIMA
-----------------
"""

import numpy as np

from statsmodels.tsa.arima.model import ARIMA

from oats.models._base import Model


class ARIMAModel(Model):
    """Autoregressive Intergrated Moving Average Model

    Implemented using statsmodels package. Multivariate scoring enabled by fitting and predicting each feature column.
    """

    def __init__(self, p=1, d=1, q=1, **kwargs):
        """
        Common parameters:
            - ARIMA(1,0,0) = first-order autoregressive model
            - ARIMA(0,1,0) = random walk
            - ARIMA(1,1,0) = differenced first-order autoregressive model
            - ARIMA(0,1,1) without constant = simple exponential smoothing
            - ARIMA(0,1,1) with constant = simple exponential smoothing with growth
            - ARIMA(0,2,1) or (0,2,2) without constant = linear exponential smoothing
            - ARIMA(1,1,2) with constant = damped-trend linear exponential smoothing`

        Args:
            p (int, optional): _description_. Defaults to 1.
            d (int, optional): _description_. Defaults to 1.
            q (int, optional): _description_. Defaults to 1.
        """
        self.order = (p, d, q)

    def fit(self, train, *args, **kwargs):
        if train.ndim > 1 and train.shape[1] > 1:
            self._models = self._pseudo_mv_train(train)
            return

        self.model = ARIMA(train, order=self.order)
        self.fitted = self.model.fit()

    def get_scores(self, data):
        if data.ndim > 1 and data.shape[1] > 1:
            return self._handle_multivariate(data, self._models)

        # test data has to be immediately after train
        scores = np.zeros(len(data))
        fitted = self.fitted
        for i in range(len(data)):
            forecast = fitted.forecast(1)
            scores[i] = forecast - data[i]

            fitted = fitted.append([data[i]], refit=False)

        return np.abs(scores)



"""
Peaks-Over-Threshold (POT)
-----------------
"""

import numpy as np
from scipy.stats import genpareto

from oats.threshold._base import Threshold


class POTThreshold(Threshold):
    """Fit the tails of the data with Generalized Pareto Distribution (GPD).
    Find the threshold where `P(thres) < q`.
    Usual values for q is 1e-3 to 1e-6.

    Siffer, Alban and Fouque, Pierre-Alain and Termier, Alexandre and Largouet, Christine
    "Anomaly Detection in Streams with Extreme Value Theory"
    https://doi.org/10.1145/3097983.3098144
    """

    @classmethod
    def _set_initial_threshold(cls, tail_level: float, scores) -> float:
        return np.quantile(scores, tail_level)

    @classmethod
    def _get_peak_set(cls, t: float, scores):
        x = scores.copy()
        x = x[x >= t]

        if len(x) == 0:
            t *= 0.95
            t = 0 if t < 1e-3 else t
            return cls._get_peak_set(t, scores)

        return x - t

    @classmethod
    def _get_gpd_param(cls, peak_set):
        mu, sigma, gamma = genpareto.fit(peak_set)
        return sigma, gamma

    def __init__(self, **kwargs):
        self._thresholders = None

    def fit(self, data):
        multivar = True if data.ndim > 1 and data.shape[1] > 1 else False
        if multivar:
            self._thresholders = self._pseudo_mv_fit(data)
            return
        return

    def get_threshold(self, data, q: float = 1e-4, tail_level: float = 0.95):
        """
        Args:
            data (np.ndarray): array of data/anomaly scores
            q (float, optional): q level such that `P(threshold) < q`. Defaults to 1e-4.
            tail_level (float, optional): threshold to fit tail distribution. Defaults to 0.95.

        Returns:
            np.ndarray: array of thresholds
        """
        multivar = True if data.ndim > 1 and data.shape[1] > 1 else False
        if multivar:
            if not self._thresholders:
                self._thresholders = self._pseudo_mv_fit(data)
            return self._handle_multivariate(
                data, self._thresholders, q=q, tail_level=tail_level
            )

        t = self._set_initial_threshold(tail_level, data)
        y = self._get_peak_set(t, data)
        n_y = len(y)
        n = len(data)
        sigma, gamma = self._get_gpd_param(y)

        new_threshold = t + sigma / gamma * ((q * n / n_y) ** (-gamma) - 1)
        return np.tile(new_threshold, len(data))



"""
N-HiTS
-----------------
"""
from typing import Any
from functools import partial

from darts import models
import numpy as np
import numpy.typing as npt
import optuna

from oats.models._darts_model import DartsModel


class NHiTSModel(DartsModel):
    """N-HiTS Model (Similar to N-BEATS but faster)

    Using N-HiTS as a predictor. Anomalies scores are deviations from predictions.

    Reference: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.nhits.html
    """

    def __init__(
        self,
        window: int = 10,
        n_steps: int = 1,
        use_gpu: bool = False,
        val_split: float = 0.2,
        **kwargs
    ):
        """
        initialization also accepts any parameters used by: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.nhits.html

        Args:
            window (int, optional): rolling window size to feed into the predictor. Defaults to 10.
            n_steps (int, optional): number of steps to predict forward. Defaults to 1.
            use_gpu (bool, optional): whether to use GPU. Defaults to False.
            val_split (float, optional): proportion of data points reserved for validation. Defaults to 0.2.
        """
        model_cls = models.NHiTSModel

        super().__init__(model_cls, window, n_steps, use_gpu, val_split, **kwargs)

    def _model_objective(self, trial, train_data: npt.NDArray[Any]):
        """
        params = {
            "num_stacks": trial.suggest_int("num_stacks", 1, 5),
            "num_blocks": trial.suggest_int("num_blocks", 1, 3),
            "num_layers": trial.suggest_int("num_layers", 1, 4),
            "dropout": trial.suggest_float("dropout", 0.0, 0.3),
            "batch_size": trial.suggest_int(
                "batch_size", 1, (len(train_data) - self.window) // self.n_steps // 4
            ),
        }
        """

        # return self._get_hyperopt_res(params, train_data)
        return 0



"""
Implementation from: https://github.com/unit8co/darts
"""

from typing import Any, Tuple
from functools import partial
import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)

import numpy as np
import numpy.typing as npt
from numpy.lib.stride_tricks import sliding_window_view
from darts.timeseries import TimeSeries
from darts.dataprocessing.transformers import Scaler
from scipy.stats import zscore
import optuna

from oats.models._base import Model


class SimpleDartsModel(Model):
    def __init__(
        self, model_cls, window: int, n_steps: int, lags: int, val_split=0.2, **kwargs
    ):
        self.window = window
        self.n_steps = n_steps
        self.lags = lags
        self.val_split = val_split
        self.val_split_mem = val_split

        self.model_cls = model_cls
        self.model = model_cls(self.lags)
        self.transformer = None
        self.params = kwargs

    @property
    def _model_name(self):
        return type(self).__name__

    def __repr__(self):
        r = {}
        r.update({"model_name": self._model_name})
        r.update({"window": self.window})
        r.update({"n_steps": self.n_steps})
        r.update({"lags": self.lags})
        r.update({"model_params": {} if not self.params else self.params})

        return str(r)

    def hyperopt_model(
        self,
        train_data: npt.NDArray[Any],
        n_trials: int = 30,
        n_jobs: int = -1,
    ):
        # TODO: we can probably merge this with the hyperparam tuning method for window size

        # obj
        obj = partial(
            self._model_objective,
            train_data=train_data,
        )

        study = optuna.create_study()
        study.optimize(obj, n_trials=n_trials, n_jobs=n_jobs)

        self.params = study.best_params

        self.model = self.model_cls(self.lags, **self.params)

    def hyperopt_ws(
        self,
        train_data: npt.NDArray[any],
        n_trials: int = 30,
        n_jobs: int = -1,
    ):
        obj = partial(
            self._ws_objective,
            train_data=train_data,
        )

        study = optuna.create_study()
        study.optimize(obj, n_trials=n_trials, n_jobs=n_jobs)

        w = study.best_params.get("w")
        s = study.best_params.get("s")
        l = study.best_params.get("l")

        self.window = w
        self.n_steps = s
        self.lags = l

        self.val_split = max(
            self.val_split_mem, (self.window + self.n_steps) / len(train_data) + 0.01
        )

        self.model = self.model_cls(l)

    def _ws_objective(
        self,
        trial,
        train_data: npt.NDArray[any],
    ):
        w_high = max(
            int(0.25 * len(train_data)), int(len(train_data) * self.val_split * 0.5)
        )

        window = trial.suggest_int("w", 20, w_high, 5)
        n_steps = trial.suggest_int("s", 1, 20)
        lags = trial.suggest_int("l", 1, 20 - 1)

        val_split = min(
            self.val_split_mem, (self.window + self.n_steps) / len(train_data) + 0.01
        )

        cls = self.__class__(window, n_steps, lags, val_split)
        cls.model = cls.model_cls(lags)
        cls.fit(train_data)

        tr, val = cls._get_train_val_split(train_data, val_split)

        try:
            _, res, _ = cls.get_scores(val)
        except ValueError:
            return 1e4

        return np.sum(res**2)

    def fit(self, train_data: npt.NDArray[Any], *args, **kwargs):
        if self.model is None:
            print("Unspecified hyperparameters, please run hyperopt_ws()")
            return

        if train_data.ndim > 1 and train_data.shape[1] > 1:
            self._models = self._pseudo_mv_train(train_data)
            return

        train_data = self._scale_series(train_data)

        tr, val = self._get_train_val_split(train_data, self.val_split)

        self.model.fit(TimeSeries.from_values(tr))

    def get_scores(self, test_data: npt.NDArray[Any]) -> Tuple[npt.NDArray[Any]]:
        # TODO: if makes sense, have a base class for this and DartsModel...

        if test_data.ndim > 1 and test_data.shape[1] > 1:
            return self._handle_multivariate(test_data, self._models)

        test_data = self._scale_series(test_data)

        windows = sliding_window_view(test_data, self.window)
        windows = windows[:: self.n_steps]

        preds = np.array([])

        seq = []
        for arr in windows.astype(np.float32):
            ts = TimeSeries.from_values(arr)
            seq.append(ts)

        scores = self.model.predict(n=self.n_steps, series=seq)

        for step in scores:
            preds = np.append(preds, step.pd_series().to_numpy())

        tdata_trim = test_data[self.window :]

        preds = preds[: len(tdata_trim)]

        residual = preds - tdata_trim
        residual = np.abs(zscore(residual))
        residual = np.append(np.zeros(self.window), residual)

        anom = np.absolute(residual)

        i_preds = TimeSeries.from_values(preds)
        i_preds = self.transformer.inverse_transform(i_preds).pd_series().to_numpy()
        i_preds = np.append(np.zeros(self.window), i_preds)

        self._preds = i_preds
        self._residual = residual

        return anom

    def _scale_series(self, series: npt.NDArray[Any]):
        series = TimeSeries.from_values(series)

        if self.transformer is None:
            self.transformer = Scaler()
            self.transformer.fit(series)

        series = self.transformer.transform(series)

        return series.pd_series().to_numpy().astype(np.float32)

    def _get_train_val_split(
        self, series: npt.NDArray[Any], pct_val: float
    ) -> Tuple[npt.NDArray[Any], npt.NDArray[Any]]:
        arr_len = len(series)
        split_at = int(arr_len * (1 - pct_val))

        return series[:split_at], series[split_at - self.window :]

    def _get_hyperopt_res(self, params: dict, train_data):
        try:
            m = self.__class__(self.window, self.n_steps, self.lags, self.val_split)
            m.model = m.model_cls(self.lags, **params)

            m.fit(train_data)

        except RuntimeError:
            return 1e4

        _, val = self._get_train_val_split(train_data, self.val_split)
        _, res, _ = m.get_scores(val)
        return np.sum(res**2)



"""
FluxEV
-----------------
"""

import time

from oats.models._base import Model
import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
from sklearn.neighbors import KernelDensity
from sklearn.model_selection import GridSearchCV
from scipy.stats import norm, genpareto


class _FluxEV2000:
    def __init__(
        self,
        window,
        q,
        level,
        alpha=0.7,
        method="all",
        p=None,
        memory=2000,
        bw_reset_every=2000,
        support=0,
        init_cutoff=1,
        mle=False,
        **kwargs,
    ):
        if method.lower() not in ["all", "above", "below"]:
            raise ValueError(
                f"method parameter expexts ['all', 'above', 'below'] but passed in {method}"
            )
        self.bw_reset_every = bw_reset_every
        self.support = support
        self.init_cutoff = init_cutoff

        self.s = window
        self.q = q
        self.level = level
        self.p = 1 if p is None else p
        self.memory = memory

        # EWMA Decay Factor
        # alpha -> 0 => normal EMA
        # alpha -> 1 => EWMA -> X[-1]
        self.alpha = alpha

        # Method: Literal["all", "above", "below"]
        # all    => admit all points
        # above  => admit only those above moving avg
        # below  => admit only those below moving average
        self.method = method

        # Raw Data
        self.X = np.zeros(window)

        # Residual after EWMA
        self.E = np.zeros(window)

        # Residual after First Smoothing
        self.F = np.zeros(p)

        # Peak Set
        self.Y = np.array([])

        # F Memory
        self.F_mem = np.array([])

        # S -> Residual after second smoothing
        self.S = np.zeros(1)

        # len(data set)
        self.n = 0

        # I -> indices of peaks in order to maintain the correct n_peak/n ratio
        self.I = np.array([0])

        # thresholds
        self.percentile_thres = 0
        self.spot_thres = 0

        # timer
        self.perf_counter = 0

        # Use MLE Estimation (more stable)
        self.mle = mle

    @property
    def adjusted_n(self):
        return self.n - self.I[0]

    # TODO: typing
    #       data -> np.NDArray
    def load_initial(self, data):
        if data.size <= 2 * self.s:
            raise ValueError(
                f"Expects initial series of length > max(2*s, p) == {max(2*self.s, self.p)}, passed in {data.size}"
            )
        t = time.perf_counter()

        self.X = data[self.s : 2 * self.s]

        s_window = sliding_window_view(data[: 2 * self.s], self.s)[:-1]
        self.E = data[self.s : 2 * self.s] - EWMA(s_window.T, self.alpha)

        Q = np.array([])
        for i in data[2 * self.s :]:
            # initialize SPOT
            self._step_train(i)
            Q = np.append(Q, self.F[-1])

        self.spot_thres = self._spot_initialize(Q)
        self.perf_counter += time.perf_counter() - t

    # TODO: typing
    #       data -> np.float / float
    def _step_train(self, x_t) -> int:
        """
        Stepping without updating SPOT
        """
        # fluctuation removal
        E_i = x_t - EWMA(self.X, self.alpha)

        if self.method == "above":
            E_i = np.maximum(0.0, E_i)
        if self.method == "below":
            E_i = np.minimum(0, E_i)

        # first smoothing
        E_new = np.append(self.E, E_i)
        E_old = self.E.copy()
        F_i = (np.std(E_new, ddof=1) - np.std(E_old, ddof=1)).clip(min=0)

        self._add_mem(F_i)

        # second smoothing
        m = np.max(self.F)
        self.S = np.where(F_i - m > 0, np.array([F_i]), np.array([0]))

        # smoothing calculations done; advance one step
        self._array_step(x_t, E_i, F_i)

    def step_predict(self, x_t) -> int:
        """
        Stepping + Updating SPOT
        """
        t = time.perf_counter()
        if self.n % self.bw_reset_every == 0:
            print(f"z_t: {self.percentile_thres} | z_q: {self.spot_thres}")
            self.bw = _SPOTMoM.fit_kde_bw(self.F_mem)

        self._step_train(x_t)
        new_p_thres = _SPOTMoM.get_tail_threshold(
            self.F_mem, self.level, samples=100, bw=self.bw
        )
        self.spot_thres += new_p_thres - self.percentile_thres
        self.percentile_thres = new_p_thres

        out = np.int32(0)
        if self.S[-1] > self.spot_thres:
            """erasing anomalous data point from history"""
            out = np.int32(1)
            self.F_mem[-1] = 0

        elif self.F[-1] > self.percentile_thres and self.F[-1] < self.spot_thres:
            # add obs to peak set
            y_i = self.F[-1] - self.percentile_thres
            self._add_peak(y_i)

            """ vanilla calc
            sigma, gamma = SPOTMoM.get_gpd_params(self.Y)
            self.spot_thres = SPOTMoM.calc_spot_threshold(self.percentile_thres, sigma, gamma, self.n, self.Y.size, self.q)
            """

            sigma, gamma = _SPOTMoM.get_gpd_params(self.Y, mle=self.mle)
            self.spot_thres = max(
                _SPOTMoM.calc_spot_threshold(
                    self.percentile_thres,
                    sigma,
                    gamma,
                    self.adjusted_n,
                    self.Y.size,
                    self.q,
                ),
                _SPOTMoM.calc_half_normal_threshold(
                    self.percentile_thres,
                    self.Y.std(ddof=1),
                    self.q,
                    support=self.support,
                ),
            )

        elif self.F[-1] > self.spot_thres:
            self.F_mem[-1] = 0

        self.perf_counter += time.perf_counter() - t
        return out

    def _array_step(self, X_i, E_i, F_i):
        self.X = np.roll(self.X, -1)
        self.X[-1] = X_i

        self.E = np.roll(self.E, -1)
        self.E[-1] = E_i

        self.F = np.roll(self.F, -1)
        self.F[-1] = F_i

        self.n += 1

    @property
    def n_peaks(self):
        return self.Y.size

    def _spot_initialize(self, S):
        # print(f"Q_len: {len(S)} | Q_non_zeros: {np.count_nonzero(S)} | Q_max: {np.max(S)}")
        # print(S[S>0])
        self.bw = _SPOTMoM.fit_kde_bw(S)
        self.percentile_thres = _SPOTMoM.get_tail_threshold(S, self.level, bw=self.bw)

        # print(f"thres: {self.percentile_thres}")
        # print(f"n_above: {((S-self.percentile_thres) > 0).sum()}")

        Y = _SPOTMoM.get_peaks(S, self.percentile_thres)
        I = _SPOTMoM.get_peaks_idx(S, self.percentile_thres)

        if self.init_cutoff < 1:
            bw = _SPOTMoM.fit_kde_bw(Y)
            cutoff = _SPOTMoM.get_tail_threshold(Y, self.init_cutoff, bw=bw)
            Y = Y[Y < cutoff]

        for y, i in zip(Y, I):
            self._add_peak(y, i)

        sigma, gamma = _SPOTMoM.get_gpd_params(Y, mle=self.mle)

        n_y = len(Y)
        n = len(S)

        spot_thres = max(
            _SPOTMoM.calc_spot_threshold(
                self.percentile_thres, sigma, gamma, n, n_y, self.q
            ),
            _SPOTMoM.calc_half_normal_threshold(
                self.percentile_thres, Y.std(ddof=1), self.q, support=self.support
            ),
        )

        return spot_thres

    def _add_peak(self, y, index=None):
        if self.Y.size <= self.memory:
            self.Y = np.append(self.Y, y)
        else:
            self.Y = np.roll(self.Y, -1)
            self.Y[-1] = y

        index = self.n if index is None else index
        if self.I.size <= self.memory + 1:
            self.I = np.append(self.I, index)
        else:
            self.I = np.roll(self.I, -1)
            self.I[-1] = index

    def _add_mem(self, f):
        if self.F_mem.size <= self.memory:
            self.F_mem = np.append(self.F_mem, f)
        else:
            self.F_mem = np.roll(self.F_mem, -1)
            self.F_mem[-1] = f


def EWMA(x, alpha, flip_x=True):
    if flip_x:
        x = np.flip(x)

    s = len(x)
    a = np.full(s, 1 - alpha)
    power = np.arange(0, s)
    weights = np.power(a, power)

    res = np.dot(weights.T, x) / np.sum(weights)
    return np.flip(res)


class _SPOTMoM:
    @classmethod
    def fit_kde_bw(cls, data, kernel="gaussian"):
        params = {"bandwidth": np.logspace(-5, 10, 100)}
        grid = GridSearchCV(KernelDensity(rtol=0.01), params, n_jobs=-1)
        grid.fit(data[:, np.newaxis])
        return grid.best_params_["bandwidth"]

    @classmethod
    def get_tail_threshold(
        cls, data, level, samples=1000, bw=0.5, kernel="gaussian", rtol=0.02
    ) -> float:
        # via KDE
        data = np.append(data, 1e-3)  # added for stability in case data is all zeros
        x = np.linspace(data.min() + 1e-3, data.max(), samples)

        kde = KernelDensity(kernel=kernel, rtol=rtol, bandwidth=bw).fit(
            data[:, np.newaxis]
        )
        pdf = np.exp(kde.score_samples(x[:, np.newaxis]))

        cdf = np.cumsum(pdf)
        cdf /= cdf.max()

        if len(np.where(cdf <= level)[0]) == 0:
            return cls.get_tail_threshold(data, level, samples * 5, bw, kernel)

        idx = np.where(cdf <= level)[0][-1]

        return x[idx]

    @classmethod
    def get_peaks(cls, data, threshold, n_sample_min=5):
        threshold = cls._adjust_threshold(data, threshold, n_sample_min)
        return data[data >= threshold] - threshold

    @classmethod
    def get_peaks_idx(cls, data, threshold, n_sample_min=5):
        threshold = cls._adjust_threshold(data, threshold, n_sample_min)
        return np.where(data >= threshold)[0]

    @classmethod
    def _adjust_threshold(cls, data, threshold, n_sample_min):
        x = data.copy()
        x = x[x >= threshold]

        if len(x) < n_sample_min:
            threshold *= 0.95
            threshold = 0 if threshold < 1e-3 else threshold
            return cls._adjust_threshold(data, threshold, n_sample_min)

        return threshold

    @classmethod
    def get_gpd_params_mv(cls, mean, var):
        if var == 0:
            return 1, 1

        sigma = mean / 2 * (1 + mean**2 / var)
        gamma = 1 / 2 * (1 - (mean**2 / var))

        return sigma, gamma

    @classmethod
    def get_gpd_params(cls, peaks, robust=False, mle=False):
        y = peaks.copy()

        if mle:
            mu, sigma, gamma = genpareto.fit(y)
            return sigma, gamma

        if robust:
            huber.maxiter = 100
            huber.tol = 5e-2
            mu, std = huber(y)
            var_y = std**2
        else:
            mu = y.mean()
            var_y = y.var(ddof=1)

        sigma = mu / 2 * (1 + mu**2 / var_y)
        gamma = 1 / 2 * (1 - (mu**2 / var_y))

        return sigma, gamma

    @classmethod
    def calc_spot_threshold(cls, initial_threshold, sigma, gamma, n, n_peaks, q):
        delta = sigma / gamma * ((q * n / n_peaks) ** (-gamma) - 1)
        new_thres = initial_threshold + delta

        return new_thres

    @classmethod
    def calc_half_normal_threshold(cls, initial_threshold, std_dev, q, support=1):
        return initial_threshold + std_dev * norm.ppf(((1 - q) + 1) / 2) * support


class FluxEVModel(Model):
    """FluxEV

    Implementation from the paper with QoL improvements for production.

    Li, Jia and Di, Shimin and Shen, Yanyan and Chen, Lei
    "FluxEV: A Fast and Effective Unsupervised Framework for Time-Series Anomaly Detection"
    https://doi.org/10.1145/3437963.3441823
    """

    def __init__(
        self, window: int = 10, window_smoothing=None, q=1e-4, level=0.95, **kwargs
    ):
        """
        Args:
            window (int, optional): main window length. Defaults to 10.
            window_smoothing (int, optional): window length for smoothig operation, if None then same as `window`. Defaults to None.
            q (float, optional): q level such that `P(threshold) < q`. Defaults to 1e-4.
            level (float, optional): threshold to fit tail distribution. Defaults to 0.95.
            memory (int, optional): how many data points for the tails. Defaults to 2000.
            bw_reset_every (int, optional): used for method of moments estimation to prevent a bad fit (MoM is not robust against outliers); use distribution from kde estimation to compute tail level; this parameter dictates how often bandwidth is refitted. Defaults to 2000.
            support (float, optional): used for method of moments estimation to prevent a bad fit (MoM is fast but unreliable); threshold is computed as the maximum of `GPD` and `support * HalfNormal`. Defaults to 0.
            init_cutoff (float, optional): used for method of moments estimation to prevent a bad fit (MoM is not robust against outliers); only bottom `init_cutoff` deciles are used to fit the GPD distribution. Defaults to 1.
            mle (bool, optional): whether to use MLE for GPD parameter estimation; setting True will result in a slow down in performance. Default to False.

        """
        self.window = window
        self.window_smoothing = window if window_smoothing is None else window_smoothing
        self.q = q
        self.level = level

        self.params = kwargs

    def fit(self, data, **kwargs):
        if data.ndim > 1 and data.shape[1] == 1:
            data = data.flatten()

        if data.ndim > 1 and data.shape[1] > 1:
            self._models = self._pseudo_mv_train(data)
            return

        self.model = _FluxEV2000(
            self.window,
            q=self.q,
            level=self.level,
            p=self.window_smoothing,
            **self.params,
        )
        self.model.load_initial(data)

    def get_scores(self, data):
        multivar = True if data.ndim > 1 and data.shape[1] > 1 else False
        if multivar:
            return self._handle_multivariate(data, self._models)

        if data.ndim > 1 and data.shape[1] == 1:
            data = data.flatten()

        anoms = np.zeros(len(data))
        for idx, value in enumerate(data):
            anoms[idx] = self.model.step_predict(value)

        return anoms



from dataclasses import dataclass, astuple
from typing import Tuple, Any, Union

import numpy as np
import numpy.typing as npt

from oats._data.base import DataReader, Data
from oats._utils.utils import array_safe_eq


@dataclass(eq=False)
class UcrData(Data):
    series: npt.NDArray[Any]
    labels: npt.NDArray[Any]
    train_len: int
    file_name: str

    @property
    def train(self) -> Tuple[npt.NDArray[Any], npt.NDArray[Any]]:
        # data, label
        tl = self.train_len
        return self.series[:tl], self.labels[:tl]

    @property
    def test(self) -> Tuple[npt.NDArray[Any], npt.NDArray[Any]]:
        # data, label
        tl = self.train_len
        return self.series[tl:], self.labels[tl:]

    def get_test_with_window(
        self, window: int
    ) -> Tuple[npt.NDArray[Any], npt.NDArray[Any]]:
        tl = self.train_len
        return self.series[tl - window :], self.labels[tl:]

    def __eq__(self, other) -> bool:
        if self is other:
            return True

        if self.__class__ is not other.__class__:
            return NotImplemented

        t1 = astuple(self)
        t2 = astuple(other)

        return all(array_safe_eq(a1, a2) for a1, a2 in zip(t1, t2))


class UcrDataReader(DataReader):
    def __init__(self):
        self.path: Union[str, None] = None

    def __call__(self, path: str) -> UcrData:
        self.set_path(path)
        return self.read()

    def set_path(self, path: str) -> None:
        self.path = path

    def read(self) -> UcrData:
        try:
            series = np.loadtxt(self.path)
        except FileNotFoundError:
            return None

        train_len = self._get_train_len()
        labels = self._get_labels(series)
        file_name = self.path.split("/")[-1]

        return UcrData(series, labels, train_len, file_name)

    def _get_labels(self, series: npt.NDArray[Any]) -> npt.NDArray[Any]:
        start_idx, end_idx = self._get_label_range()
        arr = np.zeros(series.size)
        arr[start_idx - 1 : end_idx] = 1.0

        return arr

    def _get_file_name(self) -> str:
        if self.path is None:
            return None

        return self.path.split("/")[-1].split(".")[0]

    def _get_train_len(self) -> int:
        if self.path is None:
            return None

        file_name = self._get_file_name()
        return int(file_name.split("_")[-3])

    def _get_label_range(self) -> Tuple[int, int]:
        if self.path is None:
            return None

        file_name = self._get_file_name()
        fn_split = file_name.split("_")
        return int(fn_split[-2]), int(fn_split[-1])



import torch.nn as nn
import torch
import torch.nn.functional as F
from torch.autograd import Variable
import math
import numpy as np


class ConvLSTMCell(nn.Module):
    def __init__(self, input_dim, hidden_dim, kernel_size, bias):
        """
        Initialize ConvLSTM cell.

        Parameters
        ----------
        input_dim: int
            Number of channels of input tensor.
        hidden_dim: int
            Number of channels of hidden state.
        kernel_size: (int, int)
            Size of the convolutional kernel.
        bias: bool
            Whether or not to add the bias.
        """

        super(ConvLSTMCell, self).__init__()

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim

        self.kernel_size = kernel_size
        self.padding = kernel_size[0] // 2, kernel_size[1] // 2
        self.bias = bias

        self.conv = nn.Conv2d(
            in_channels=self.input_dim + self.hidden_dim,
            out_channels=4 * self.hidden_dim,
            kernel_size=self.kernel_size,
            padding=self.padding,
            bias=self.bias,
        )

    def forward(self, input_tensor, cur_state):
        h_cur, c_cur = cur_state

        combined = torch.cat(
            [input_tensor, h_cur], dim=1
        )  # concatenate along channel axis

        combined_conv = self.conv(combined)
        cc_i, cc_f, cc_o, cc_g = torch.split(combined_conv, self.hidden_dim, dim=1)
        i = torch.sigmoid(cc_i)
        f = torch.sigmoid(cc_f)
        o = torch.sigmoid(cc_o)
        g = torch.tanh(cc_g)

        c_next = f * c_cur + i * g
        h_next = o * torch.tanh(c_next)

        return h_next, c_next

    def init_hidden(self, batch_size, image_size):
        height, width = image_size
        return (
            torch.zeros(
                batch_size,
                self.hidden_dim,
                height,
                width,
                device=self.conv.weight.device,
            ),
            torch.zeros(
                batch_size,
                self.hidden_dim,
                height,
                width,
                device=self.conv.weight.device,
            ),
        )


class ConvLSTM(nn.Module):

    """

    Parameters:
        input_dim: Number of channels in input
        hidden_dim: Number of hidden channels
        kernel_size: Size of kernel in convolutions
        num_layers: Number of LSTM layers stacked on each other
        batch_first: Whether or not dimension 0 is the batch or not
        bias: Bias or no bias in Convolution
        return_all_layers: Return the list of computations for all layers
        Note: Will do same padding.

    Input:
        A tensor of size B, T, C, H, W or T, B, C, H, W
    Output:
        A tuple of two lists of length num_layers (or length 1 if return_all_layers is False).
            0 - layer_output_list is the list of lists of length T of each output
            1 - last_state_list is the list of last states
                    each element of the list is a tuple (h, c) for hidden state and memory
    Example:
        >> x = torch.rand((32, 10, 64, 128, 128))
        >> convlstm = ConvLSTM(64, 16, 3, 1, True, True, False)
        >> _, last_states = convlstm(x)
        >> h = last_states[0][0]  # 0 for layer index, 0 for h index
    """

    def __init__(
        self,
        input_dim,
        hidden_dim,
        kernel_size,
        num_layers,
        batch_first=False,
        bias=True,
        return_all_layers=False,
    ):
        super(ConvLSTM, self).__init__()

        self._check_kernel_size_consistency(kernel_size)

        # Make sure that both `kernel_size` and `hidden_dim` are lists having len == num_layers
        kernel_size = self._extend_for_multilayer(kernel_size, num_layers)
        hidden_dim = self._extend_for_multilayer(hidden_dim, num_layers)
        if not len(kernel_size) == len(hidden_dim) == num_layers:
            raise ValueError("Inconsistent list length.")

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.kernel_size = kernel_size
        self.num_layers = num_layers
        self.batch_first = batch_first
        self.bias = bias
        self.return_all_layers = return_all_layers

        cell_list = []
        for i in range(0, self.num_layers):
            cur_input_dim = self.input_dim if i == 0 else self.hidden_dim[i - 1]

            cell_list.append(
                ConvLSTMCell(
                    input_dim=cur_input_dim,
                    hidden_dim=self.hidden_dim[i],
                    kernel_size=self.kernel_size[i],
                    bias=self.bias,
                )
            )

        self.cell_list = nn.ModuleList(cell_list)

    def forward(self, input_tensor, hidden_state=None):
        """

        Parameters
        ----------
        input_tensor: todo
            5-D Tensor either of shape (t, b, c, h, w) or (b, t, c, h, w)
        hidden_state: todo
            None. todo implement stateful

        Returns
        -------
        last_state_list, layer_output
        """
        if not self.batch_first:
            # (t, b, c, h, w) -> (b, t, c, h, w)
            input_tensor = input_tensor.permute(1, 0, 2, 3, 4)

        b, _, _, h, w = input_tensor.size()

        # Implement stateful ConvLSTM
        if hidden_state is not None:
            raise NotImplementedError()
        else:
            # Since the init is done in forward. Can send image size here
            hidden_state = self._init_hidden(batch_size=b, image_size=(h, w))

        layer_output_list = []
        last_state_list = []

        seq_len = input_tensor.size(1)
        cur_layer_input = input_tensor

        for layer_idx in range(self.num_layers):

            h, c = hidden_state[layer_idx]
            output_inner = []
            for t in range(seq_len):
                h, c = self.cell_list[layer_idx](
                    input_tensor=cur_layer_input[:, t, :, :, :], cur_state=[h, c]
                )
                output_inner.append(h)

            layer_output = torch.stack(output_inner, dim=1)
            cur_layer_input = layer_output

            layer_output_list.append(layer_output)
            last_state_list.append([h, c])

        if not self.return_all_layers:
            layer_output_list = layer_output_list[-1:]
            last_state_list = last_state_list[-1:]

        return layer_output_list, last_state_list

    def _init_hidden(self, batch_size, image_size):
        init_states = []
        for i in range(self.num_layers):
            init_states.append(self.cell_list[i].init_hidden(batch_size, image_size))
        return init_states

    @staticmethod
    def _check_kernel_size_consistency(kernel_size):
        if not (
            isinstance(kernel_size, tuple)
            or (
                isinstance(kernel_size, list)
                and all([isinstance(elem, tuple) for elem in kernel_size])
            )
        ):
            raise ValueError("`kernel_size` must be tuple or list of tuples")

    @staticmethod
    def _extend_for_multilayer(param, num_layers):
        if not isinstance(param, list):
            param = [param] * num_layers
        return param


class PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model).float() * (-math.log(10000.0) / d_model)
        )
        pe += torch.sin(position * div_term)
        pe += torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer("pe", pe)

    def forward(self, x, pos=0):
        x = x + self.pe[pos : pos + x.size(0), :]
        return self.dropout(x)


class TransformerEncoderLayer(nn.Module):
    def __init__(self, d_model, nhead, dim_feedforward=16, dropout=0):
        super(TransformerEncoderLayer, self).__init__()
        self.self_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.activation = nn.LeakyReLU(True)

    def forward(self, src, src_mask=None, src_key_padding_mask=None, **kwargs):
        src2 = self.self_attn(src, src, src)[0]
        src = src + self.dropout1(src2)
        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout2(src2)
        return src


class TransformerDecoderLayer(nn.Module):
    def __init__(self, d_model, nhead, dim_feedforward=16, dropout=0):
        super(TransformerDecoderLayer, self).__init__()
        self.self_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)
        self.multihead_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.dropout3 = nn.Dropout(dropout)

        self.activation = nn.LeakyReLU(True)

    def forward(
        self,
        tgt,
        memory,
        tgt_mask=None,
        memory_mask=None,
        tgt_key_padding_mask=None,
        memory_key_padding_mask=None,
        **kwargs
    ):
        tgt2 = self.self_attn(tgt, tgt, tgt)[0]
        tgt = tgt + self.dropout1(tgt2)
        tgt2 = self.multihead_attn(tgt, memory, memory)[0]
        tgt = tgt + self.dropout2(tgt2)
        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
        tgt = tgt + self.dropout3(tgt2)
        return tgt


class ComputeLoss:
    def __init__(self, model, lambda_energy, lambda_cov, device, n_gmm):
        self.model = model
        self.lambda_energy = lambda_energy
        self.lambda_cov = lambda_cov
        self.device = device
        self.n_gmm = n_gmm

    def forward(self, x, x_hat, z, gamma):
        """Computing the loss function for DAGMM."""
        reconst_loss = torch.mean((x - x_hat).pow(2))

        sample_energy, cov_diag = self.compute_energy(z, gamma)

        loss = (
            reconst_loss
            + self.lambda_energy * sample_energy
            + self.lambda_cov * cov_diag
        )
        return Variable(loss, requires_grad=True)

    def compute_energy(self, z, gamma, phi=None, mu=None, cov=None, sample_mean=True):
        """Computing the sample energy function"""
        if (phi is None) or (mu is None) or (cov is None):
            phi, mu, cov = self.compute_params(z, gamma)

        z_mu = z.unsqueeze(1) - mu.unsqueeze(0)

        eps = 1e-12
        cov_inverse = []
        det_cov = []
        cov_diag = 0
        for k in range(self.n_gmm):
            cov_k = cov[k] + (torch.eye(cov[k].size(-1)) * eps).to(self.device)
            cov_inverse.append(torch.inverse(cov_k).unsqueeze(0))
            det_cov.append(
                (Cholesky.apply(cov_k.cpu() * (2 * np.pi)).diag().prod()).unsqueeze(0)
            )
            cov_diag += torch.sum(1 / cov_k.diag())

        cov_inverse = torch.cat(cov_inverse, dim=0)
        det_cov = torch.cat(det_cov).to(self.device)

        E_z = -0.5 * torch.sum(
            torch.sum(z_mu.unsqueeze(-1) * cov_inverse.unsqueeze(0), dim=-2) * z_mu,
            dim=-1,
        )
        E_z = torch.exp(E_z)
        E_z = -torch.log(
            torch.sum(
                phi.unsqueeze(0) * E_z / (torch.sqrt(det_cov)).unsqueeze(0), dim=1
            )
            + eps
        )
        if sample_mean == True:
            E_z = torch.mean(E_z)
        return E_z, cov_diag

    def compute_params(self, z, gamma):
        """Computing the parameters phi, mu and gamma for sample energy function"""
        # K: number of Gaussian mixture components
        # N: Number of samples
        # D: Latent dimension
        # z = NxD
        # gamma = NxK

        # phi = D
        phi = torch.sum(gamma, dim=0) / gamma.size(0)

        # mu = KxD
        mu = torch.sum(z.unsqueeze(1) * gamma.unsqueeze(-1), dim=0)
        mu /= torch.sum(gamma, dim=0).unsqueeze(-1)

        z_mu = z.unsqueeze(1) - mu.unsqueeze(0)
        z_mu_z_mu_t = z_mu.unsqueeze(-1) * z_mu.unsqueeze(-2)

        # cov = K x D x D
        cov = torch.sum(gamma.unsqueeze(-1).unsqueeze(-1) * z_mu_z_mu_t, dim=0)
        cov /= torch.sum(gamma, dim=0).unsqueeze(-1).unsqueeze(-1)

        return phi, mu, cov


class Cholesky(torch.autograd.Function):
    def forward(ctx, a):
        l = torch.cholesky(a, False)
        ctx.save_for_backward(l)
        return l

    def backward(ctx, grad_output):
        (l,) = ctx.saved_variables
        linv = l.inverse()
        inner = torch.tril(torch.mm(l.t(), grad_output)) * torch.tril(
            1.0 - Variable(l.data.new(l.size(1)).fill_(0.5).diag())
        )
        s = torch.mm(linv.t(), torch.mm(inner, linv))
        return s



"""
Quantile
-----------------
"""
import numpy as np

from oats.threshold._base import Threshold


class QuantileThreshold(Threshold):
    """Simple threshold method where top `q` deciles of anomaly scores are labeled as anomalies.
    Beware that user is guaranteed to have outlier predictions with this method!

    `fit()` not necessary.
    """

    def fit(self, *args, **kwargs):
        return

    def get_threshold(self, data, percentile: float = 0.95):
        """
        Args:
            data (np.ndarray): array of anomaly scores
            percentile (float, optional): decile level used for threshold. Defaults to 0.95.

        Returns:
            np.ndarray: array of thresholds
        """
        multivar = True if data.ndim > 1 and data.shape[1] > 1 else False
        if multivar:
            tile = (len(data), 1)
        else:
            tile = len(data)

        return np.tile(np.quantile(data, percentile, axis=0), tile)



from typing import Protocol, Tuple, Any

import numpy as np
import numpy.typing as npt


class Data(Protocol):
    series: npt.NDArray[Any]
    labels: npt.NDArray[Any]
    train_len: int

    @property
    def train(self) -> Tuple[npt.NDArray[Any], npt.NDArray[Any]]:
        raise NotImplementedError

    @property
    def test(self) -> Tuple[npt.NDArray[Any], npt.NDArray[Any]]:
        raise NotImplementedError

    def __eq__(self, other) -> bool:
        raise NotImplementedError


class DataReader(Protocol):
    def __call__(self, path: str) -> Data:
        raise NotImplementedError

    def read(self) -> Data:
        raise NotImplementedError



"""
Moving Average
-----------------
"""

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
from scipy.stats import zscore

from oats.models._base import Model


class MovingAverageModel(Model):
    """Moving Average Model

    Using the average of past `window` values as a predictor for current. Anomalies scores are deviations from predictions.
    """

    def __init__(self, window: int = 10):
        """
        Args:
            window (int, optional): size of the rolling window used to past moving average. Defaults to 10.
        """
        self.window = window

    def fit(self, *args, **kwargs):
        return

    def get_scores(self, data, normalize=False):
        # if multivariate
        if data.ndim > 1 and data.shape[1] > 1:
            return self._handle_multivariate(data, [self] * data.shape[1])

        if data.ndim > 1 and data.shape[1] == 1:
            data = data.flatten()

        E = np.zeros(len(data))
        s_window = sliding_window_view(data, self.window)[:-1]

        E[self.window :] = data[self.window :] - np.mean(s_window, axis=1)

        if normalize:
            E = zscore(E)

        return np.abs(E)



"""
Implementation from: https://github.com/unit8co/darts
"""
from typing import Any, Tuple
from functools import partial

import numpy as np
import numpy.typing as npt
from numpy.lib.stride_tricks import sliding_window_view
from darts.timeseries import TimeSeries
from darts.dataprocessing.transformers import Scaler
from torch.cuda import device_count
from scipy.stats import zscore
import optuna

from oats.models._base import Model
from oats._utils.utils import get_default_early_stopping


class DartsModel(Model):
    _support_multivariate = True

    def __init__(
        self,
        model_cls,
        window: int,
        n_steps: int,
        use_gpu: bool,
        val_split: float = 0.2,
        rnn_model: str = None,
        **kwargs,
    ):

        self.window = window
        self.n_steps = n_steps
        self.use_gpu = use_gpu
        self.val_split = val_split
        self.val_split_mem = val_split
        self.rnn_model = rnn_model

        self.model_cls = model_cls

        # TODO: maybe seperate these into another class? fine for now...
        self.transformer = None

        self.params = None

        self._init_model(**kwargs)

    @property
    def _model_name(self):
        name = type(self).__name__
        if self.rnn_model:
            return f"{name}_{self.rnn_model}"

        return name

    def __repr__(self):
        r = {}
        r.update({"model_name": self._model_name})
        r.update({"window": self.window})
        r.update({"n_steps": self.n_steps})
        r.update({"val_split": self.val_split})
        r.update({"model_params": {} if not self.params else self.params})

        return str(r)

    def _init_model(self, **kwargs):
        if self.rnn_model:
            self.model = self.model_cls(
                self.window,
                training_length=self.window,
                pl_trainer_kwargs=self._get_trainer_kwargs(),
                model=self.rnn_model,
                **kwargs,
            )
            return

        self.model = self.model_cls(
            self.window,
            self.n_steps,
            pl_trainer_kwargs=self._get_trainer_kwargs(),
            **kwargs,
        )

    def _get_trainer_kwargs(self) -> dict:
        d = {}

        if self.use_gpu:
            d.update({"accelerator": "gpu", "gpus": [i for i in range(device_count())]})

        d.update({"callbacks": [get_default_early_stopping()]})
        return d

    def hyperopt_model(
        self,
        train_data: npt.NDArray[Any],
        n_trials: int = 30,
        n_jobs: int = 1,
    ):
        # TODO: we can probably merge this with the hyperparam tuning method for window size

        obj = partial(
            self._model_objective,
            train_data=train_data,
        )

        study = optuna.create_study()
        study.optimize(obj, n_trials=n_trials, n_jobs=n_jobs)

        self.params = study.best_params
        self._init_model(**self.params)

    def hyperopt_ws(
        self,
        train_data: npt.NDArray[any],
        n_trials: int = 30,
        n_jobs: int = 1,
    ):
        obj = partial(
            self._ws_objective,
            train_data=train_data,
        )

        study = optuna.create_study()
        study.optimize(obj, n_trials=n_trials, n_jobs=n_jobs)

        w = study.best_params.get("w")
        s = study.best_params.get("s")

        self.window = w
        self.n_steps = s
        self.val_split = max(
            self.val_split_mem, (self.window + self.n_steps) / len(train_data) + 0.01
        )

        self._init_model()

    def _ws_objective(
        self,
        trial,
        train_data: npt.NDArray[any],
    ):
        w_high = min(
            int(0.25 * len(train_data)), int(len(train_data) * self.val_split * 0.5)
        )

        window = trial.suggest_int("w", 20, w_high, 5)
        n_steps = trial.suggest_int("s", 1, 20)

        val_split = max(self.val_split_mem, (window + n_steps) / len(train_data) + 0.01)

        try:
            m = self.__class__(window, n_steps, self.use_gpu, val_split)
            m.fit(train_data)
        except RuntimeError:
            return 1e4

        _, val = self._get_train_val_split(train_data, self.val_split)

        try:
            _, res, _ = m.get_scores(val)
        except ValueError:
            return 1e4

        return np.sum(res**2)

    def fit(self, train_data: npt.NDArray[Any], epochs: int = 15, **kwargs):

        train_data = self._scale_series(train_data)
        tr, val = self._get_train_val_split(train_data, self.val_split)

        self.model.fit(
            TimeSeries.from_values(tr),
            val_series=TimeSeries.from_values(val),
            epochs=epochs,
            num_loader_workers=1,
            **kwargs,
        )

    def get_scores(
        self, test_data: npt.NDArray[np.float32], normalize=False, **kwargs
    ) -> Tuple[npt.NDArray[np.float32]]:
        test_data = self._scale_series(test_data)

        windows = sliding_window_view(test_data, self.window, axis=0)
        windows = windows[:: self.n_steps]
        multivar = True if test_data.ndim > 1 and test_data.shape[1] > 1 else False

        if test_data.ndim > 1 and test_data.shape[1] == 1:
            test_data = test_data.flatten()

        preds = np.ndarray(windows.shape[1]) if multivar else np.array([])

        seq = []
        for arr in windows.astype(np.float32):
            if multivar:
                arr = arr.T
            else:
                arr = arr.flatten()

            ts = TimeSeries.from_values(arr)
            seq.append(ts)

        scores = self.model.predict(n=self.n_steps, series=seq, **kwargs)

        for step in scores:
            if multivar:
                preds = np.vstack((preds, step.pd_dataframe().to_numpy()))
            else:
                preds = np.append(preds, step.pd_dataframe().to_numpy())

        tdata_trim = test_data[self.window :]

        preds = preds[: len(tdata_trim)]

        residual = preds - tdata_trim

        if normalize:
            residual = zscore(residual)

        residual = np.abs(residual)

        if multivar:
            residual = np.append(
                np.zeros((self.window, test_data.shape[1])), residual, axis=0
            )
        else:
            residual = np.append(np.zeros(self.window), residual)

        anom = np.absolute(residual)

        i_preds = TimeSeries.from_values(preds[: len(tdata_trim)])
        i_preds = self.transformer.inverse_transform(i_preds).pd_dataframe().to_numpy()

        self._preds = i_preds
        self._residual = residual

        return anom

    def _get_train_val_split(
        self, series: npt.NDArray[Any], pct_val: float
    ) -> Tuple[npt.NDArray[Any], npt.NDArray[Any]]:
        arr_len = len(series)
        split_at = int(arr_len * (1 - pct_val))

        return series[:split_at], series[split_at - self.window :]

    def _scale_series(self, series: npt.NDArray[Any]):
        series = TimeSeries.from_values(series)

        if self.transformer is None:
            self.transformer = Scaler()
            self.transformer.fit(series)

        series = self.transformer.transform(series)

        return series.pd_dataframe().to_numpy().astype(np.float32)

    def _get_hyperopt_res(self, params: dict, train_data):
        try:
            m = self.__class__(self.window, self.n_steps, self.use_gpu, self.val_split)
            m._init_model(**params)
            m.fit(train_data)

        except RuntimeError:
            return 1e4

        _, val = self._get_train_val_split(train_data, self.val_split)
        _, res, _ = m.get_scores(val)
        return np.sum(res**2)



class Scorer():
    """Base class for Scorers.
    Scorer computes the result of a predictor; can be either unsupervised or supervised.

    Under supervised setting, predicted labels are compared against actual labels to compute traditional metrics such as `f1`.
    Under unsupervised setting, anomaly scores and/or predicted labels are usually used in conjunction with original data to compute metrics.

    Example:
        # get anomaly scores
        anom_score = model.get_scores()

        # get threshold
        thresholder = QuantileThreshold()
        threshold = thresholder.get_threshold(anom_score, 0.95)

        preds = anom_score > threshold

        # Scoring output
        scorer = SupervisedScorer()
        scorer.process(preds, labels)

        print(f"F1: {scorer.f1}")
        print(f"Precision: {scorer.precision}")
        print(f"Recall: {scorer.recall}")
    """

    def process(self, *args, **kwargs):
        raise NotImplementedError



from typing import Any

import numpy as np


class Threshold():
    """Base class for thresholders

    Preprocessor object must take a `fit()` method as well as `get_threshold()`.
    If no fitting is required, concrete classes can simply return when `fit()` is called.

    Shape of returned array must be:
        Univariate: (t, ) & (t, 1) --> (t, )
        Multivariate: (t, n) --> (t, n)
    """

    def fit(self):
        raise NotImplementedError

    def get_threshold(self):
        raise NotImplementedError

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.get_threshold(*args, **kwargs)

    def _handle_multivariate(self, data, thresholders, **kwargs):
        scores = []
        for idx, series in enumerate(data.T):
            print(series.shape, len(series.shape))
            scores.append(thresholders[idx].get_threshold(series, **kwargs))

        return np.array(scores).T

    def _pseudo_mv_fit(self, data, **kwargs):
        thresholders = []
        for _ in range(data.shape[1]):
            params = self.__dict__
            thresholders.append(self.__class__(**params))

        for idx, series in enumerate(data.T):
            thresholders[idx].fit(series, **kwargs)

        return thresholders



"""
Matrix Profile
-----------------
"""

from oats.models._base import Model
from stumpy import stump, scrump, mstump, gpu_stump
import numpy as np


class MatrixProfileModel(Model):
    def __init__(self, window: int = 10, use_gpu: bool = False, approx: bool = True):
        self.window = window
        self.use_gpu = use_gpu
        self.approx = approx

    def fit(self, *args, **kwargs):
        return

    def get_scores(self, data):
        multivar = True if data.ndim > 1 and data.shape[1] > 1 else False

        # univariate in 2-D matrix
        if data.ndim > 1 and data.shape[1] == 1:
            data = data.flatten()

        if multivar:
            model = mstump
            data = data.T
            get_scores = lambda arr: arr[0].T

            scores = model(data, self.window)
        elif self.use_gpu:
            model = gpu_stump
            get_scores = lambda arr: arr[:, 0]

            scores = model(data, self.window)
        else:
            if self.approx:
                model = scrump
                get_scores = lambda arr: arr

                scores = model(
                    data, self.window, percentage=0.01, pre_scrump=True, s=None
                )
                scores.update()
                scores = scores.P_

            else:
                model = stump
                get_scores = lambda arr: arr[:, 0]

                scores = model(data, self.window)

        scores = get_scores(scores)

        if multivar:
            scores = np.append(
                np.zeros((self.window - 1, data.T.shape[1])), scores, axis=0
            )
        else:
            scores = np.append(np.zeros(self.window - 1), scores)

        return scores



"""
Variational Auto-Encoder (VAE)
-----------------
"""

from pyod.models.vae import VAE
from oats.models._pyod_model import PyODModel


class VAEModel(PyODModel):
    """VAE Model (Variational Auto-encover)

    Using reconstruction error of the trained encoder-decoder network as anomaly scores.

    Reference: https://pyod.readthedocs.io/en/latest/_modules/pyod/models/vae.html
    """

    def __init__(self, window=10, **kwargs):
        """
        initialization also accepts any parameters used by: https://pyod.readthedocs.io/en/latest/_modules/pyod/models/vae.html

        Args:
            window (int, optional): _description_. Defaults to 10.
        """
        model_cls = VAE
        self.window = window

        super().__init__(model_cls, window, **kwargs)

    def fit(self, train_data, **kwargs):
        n_feat = (
            train_data.shape[1]
            if train_data.ndim > 1 and train_data.shape[1] > 1
            else 1
        )

        if not self.params.get("encoder_neurons"):
            self.params["encoder_neurons"] = [
                n_feat * self.window,
                n_feat * self.window // 2,
                n_feat * self.window // 4,
            ]
        if not self.params.get("decoder_neurons"):
            self.params["decoder_neurons"] = [
                n_feat * self.window // 4,
                n_feat * self.window // 2,
                n_feat * self.window,
            ]
        self.model = VAE(**self.params)

        super().fit(train_data, **kwargs)



import numpy as np


class Preprocessor():
    """Base class for Preprocessors.

    Preprocessors are any function that transforms timeseries T to T'.

    Preprocessor object must take a `fit()` method as well as `transform()`.
    If no fitting is required, concrete classes can simply return when `fit()` is called.

    Shape of returned array must be:
        Univariate: (t, ) & (t, 1) --> (t, )
        Multivariate: (t, n) --> (t, n)

    Validity of return shape can be tested by adding concrete class in `/test/test_preprocessor.py`

    Example:
        pproc = SpectralResidual()
        transformed_train = pproc.transform(train)
    """

    def fit(self, data, *args, **kwargs):
        raise NotImplementedError

    def transform(self, data, *args, **kwargs):
        raise NotImplementedError

    def _handle_multivariate(self, data, processors):
        scores = []
        for idx, series in enumerate(data.T):
            scores.append(processors[idx].transform(series))

        return np.array(scores).T

    def _pseudo_mv_fit(self, data):
        processors = []
        for _ in range(data.shape[1]):
            processors.append(self.__class__(**self.__dict__))

        for idx, series in enumerate(data.T):
            processors[idx].fit(series)

        return processors



from os import listdir
from os.path import isfile, join, isdir
from typing import Any

import numpy as np
import numpy.typing as npt
import matplotlib.pyplot as plt
from pytorch_lightning.callbacks.early_stopping import EarlyStopping


def get_files_from_path(path: str):
    return sorted([f for f in listdir(path) if isfile(join(path, f))])


def get_dirs_from_path(path: str):
    return sorted([f for f in listdir(path) if isdir(join(path, f))])


def array_safe_eq(a, b) -> bool:
    """Check if a and b are equal, even if they are numpy arrays"""
    if a is b:
        return True
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        return a.shape == b.shape and (a == b).all()
    try:
        return a == b
    except TypeError:
        return NotImplemented


def get_default_early_stopping():
    return EarlyStopping(monitor="val_loss", patience=5, min_delta=0.0001, mode="min")


def save_model_output(series: npt.NDArray[Any], model, fname: str):
    spec = str(model)
    np.savetxt(fname, series, header=spec)



"""
Implementation from: https://github.com/yzhao062/pyod
"""
from typing import Any

import numpy as np
import numpy.typing as npt
from darts.timeseries import TimeSeries
from darts.dataprocessing.transformers import Scaler
from scipy.stats import zscore
from numpy.lib.stride_tricks import sliding_window_view

from oats.models._base import Model


class PyODModel(Model):
    def __init__(self, model_cls, window: int = 10, **kwargs):
        self.window = window
        self.params = kwargs

        self.model_cls = model_cls
        self.model = model_cls(**self.params)
        self.scaler = None

    @property
    def _model_name(self):
        return type(self).__name__

    def __repr__(self):
        r = {}
        r.update({"model_name": self._model_name})
        r.update({"window": self.window})

        return str(r)

    def fit(self, train_data: npt.NDArray[Any], **kwargs):
        if "epochs" in kwargs and hasattr(self, "IS_DL") and self.IS_DL:
            self.params["epochs"] = kwargs["epochs"]
            self.model = self.model_cls(**self.params)

        windows = self._get_window(train_data)
        self.model.fit(windows)

    def get_scores(self, test_data: npt.NDArray[Any], normalize=False):
        # Multivar
        multivar = True if test_data.ndim > 1 and test_data.shape[1] > 1 else False

        windows = self._get_window(test_data)
        scores = self.model.decision_function(windows)

        if normalize:
            scores = zscore(scores)
        scores = np.abs(scores)

        scores = np.append(np.zeros(self.window - 1), scores)
        if multivar:
            scores = np.tile(scores, (test_data.shape[1], 1)).T

        return scores

    def _get_window(self, data):
        # Univariate in 2-D
        data = self._scale_series(data)
        if data.ndim > 1 and data.shape[1] == 1:
            data = data.flatten()

        # Multivar
        multivar = True if data.ndim > 1 and data.shape[1] > 1 else False

        windows = sliding_window_view(data, self.window, axis=0)

        if multivar:
            n_samples, feats, n_t = windows.shape
            windows = windows.reshape(n_samples, n_t * feats)
        return windows

    def _scale_series(self, series):
        series = TimeSeries.from_values(series)

        if self.scaler is None:
            self.scaler = Scaler()
            self.scaler.fit(series)

        series = self.scaler.transform(series)

        return series.pd_dataframe().to_numpy().astype(np.float32)



import numpy.typing as npt
import numpy as np


class Model():
    """Base class for Models

    Preprocessors are any function that transforms timeseries T to T'.

    Model object must take a `fit()` method as well as `get_scores()`.
    If no fitting is required, concrete classes can simply return when `fit()` is called.

    Shape of returned array must be:
        Univariate: (t, ) & (t, 1) --> (t, )
        Multivariate: (t, n) --> (t, n)

    Validity of return shape can be tested by adding concrete class in `/test/test_scorer.py`

    Example:
        pproc = SpectralResidual()
        transformed_train = pproc.transform(train)

    """

    def fit(self):
        raise NotImplementedError

    def get_scores(self):
        raise NotImplementedError

    @property
    def _model_name(self):
        name = type(self).__name__
        if self.rnn_model:
            return f"{name}_{self.rnn_model}"

        return name

    def _handle_multivariate(self, data, models):
        scores = []
        for idx, series in enumerate(data.T):
            scores.append(models[idx].get_scores(series))

        return np.array(scores).T

    def _pseudo_mv_train(self, data):
        models = []
        for _ in range(data.shape[1]):
            params = self.__dict__
            if "model_cls" in params:
                params.pop("model_cls")
            models.append(self.__class__(**params))

        for idx, series in enumerate(data.T):
            models[idx].fit(series)

        return models



"""
Supervised Metrics
-----------------
"""

import numpy as np

from oats.scorer._base import Scorer


class SupervisedScorer(Scorer):
    """Scorer for traditional supervised metrics; only useful if actual anomalies are known"""

    def __init__(self, delay: int = None):
        """
        Args:
            delay (int, optional): For pattern anomalies, how much tolerance to give for prediction; e.g. delay of 10 means only those predictions in the first 10 time steps of a pattern anomalies are counted as success. None if no delay needed. Defaults to None.
        """
        self.tp = 0
        self.fp = 0
        self.tn = 0
        self.fn = 0

        self.delay = delay

    def process(self, preds, labels):
        preds = preds.copy()
        labels = labels.copy()

        ground_truth_ones = np.where(labels == 1)[0]
        pred_ones = np.where(preds == 1)[0]
        ranges = self._consecutive(ground_truth_ones)

        tp, fp, tn, fn = 0, 0, 0, 0

        for idx, r in enumerate(ranges):
            intersect = np.intersect1d(r, pred_ones, assume_unique=True)
            # if alert delay more than 100 timesteps, count that as bad!

            if intersect.size != 0:
                cond = (
                    intersect[0] < r[0] + self.delay if self.delay is not None else True
                )
                if cond:
                    tp += r.size
                else:
                    fn += r.size

                preds[intersect] = 0
                pred_ones = np.where(preds == 1)[0]
            else:
                fn += r.size

        fp += pred_ones.size
        tn += preds.size - tp - fp - fn

        self.tp += tp
        self.fp += fp
        self.tn += tn
        self.fn += fn

    def _consecutive(self, data, stepsize=1):
        return np.split(data, np.where(np.diff(data) != stepsize)[0] + 1)

    @property
    def tpr(self):
        """True Positive Rate"""

        return self.tp / (self.fn + self.tp)

    @property
    def fpr(self):
        """False Positive Rate"""
        return self.fp / (self.tn + self.fp)

    @property
    def tnr(self):
        """True Negative Rate"""
        if self.tn + self.fp == 0:
            return 0
        return self.tn / (self.tn + self.fp)

    @property
    def fnr(self):
        """False Negative Rate"""
        if self.fn + self.tp == 0:
            return 0
        return self.fn / (self.fn + self.tp)

    @property
    def precision(self):
        """Precision"""
        if self.tp + self.fp == 0:
            return 0
        return self.tp / (self.tp + self.fp)

    @property
    def recall(self):
        """Recall"""
        if self.tp + self.fn == 0:
            return 0
        return self.tp / (self.tp + self.fn)

    @property
    def f1(self):
        """F-1 Score"""
        if self.recall + self.precision == 0:
            return 0
        return (2 * self.precision * self.recall) / (self.precision + self.recall)

    def __str__(self):
        return f"{self.tp}, {self.fp}, {self.tn}, {self.fn}, {self.tpr}, {self.fpr}, {self.tnr}, {self.fnr}, {self.precision}, {self.recall}, {self.f1}"



"""
N-BEATS
-----------------
"""
from typing import Any
from functools import partial

import numpy as np
import numpy.typing as npt
import optuna
from darts import models

from oats.models._darts_model import DartsModel


class NBEATSModel(DartsModel):
    """N-BEATS Model (Neural Basis Expansion Analysis Time Series Forecasting)

    Using N-BEATS as a predictor. Anomalies scores are deviations from predictions.

    Reference: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.nbeats.html
    """

    def __init__(
        self,
        window: int = 10,
        n_steps: int = 1,
        use_gpu: bool = False,
        val_split: float = 0.2,
        **kwargs
    ):
        """
        initialization also accepts any parameters used by: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.nbeats.html

        Args:
            window (int, optional): rolling window size to feed into the predictor. Defaults to 10.
            n_steps (int, optional): number of steps to predict forward. Defaults to 1.
            use_gpu (bool, optional): whether to use GPU. Defaults to False.
            val_split (float, optional): proportion of data points reserved for validation. Defaults to 0.2.
        """

        model = models.NBEATSModel

        super().__init__(model, window, n_steps, use_gpu, val_split, **kwargs)

    def _model_objective(self, trial, train_data: npt.NDArray[Any]):
        params = {
            "num_blocks": trial.suggest_int("num_blocks", 1, 2),
            "num_stacks": trial.suggest_int("num_stacks", 2, 32),
            "num_layers": trial.suggest_int("num_layers", 1, 16),
            "layer_widths": trial.suggest_int("layer_widths", 128, 512),
            "expansion_coefficient_dim": trial.suggest_int(
                "expansion_coefficient_dim", 1, 10
            ),
            "batch_size": trial.suggest_int(
                "batch_size", 1, (len(train_data) - self.window) // self.n_steps // 4
            ),
        }

        return self._get_hyperopt_res(params, train_data)



"""
Regression
-----------------
"""
from darts import models

from oats.models._darts_simple import SimpleDartsModel


class RegressionModel(SimpleDartsModel):
    """Linear Regression Model

    Using linear regression as a predictor. Anomalies scores are deviations from predictions.

    Reference: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.regression.html
    """

    def __init__(
        self,
        window: int = 10,
        n_steps: int = 1,
        lags: int = 1,
        val_split: float = 0.2,
        **kwargs
    ):
        """
        initialization also accepts any parameters used by: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.regression.html

        Args:
            window (int, optional): rolling window size to feed into the predictor. Defaults to 10.
            n_steps (int, optional): number of steps to predict forward. Defaults to 1.
            lags (int, optional): number of lags. Defaults to 1.
            val_split (float, optional): proportion of data points reserved for validation; only used if using auto-tuning (not tested). Defaults to 0.
        """
        model_cls = models.RegressionModel

        super().__init__(model_cls, window, n_steps, lags, val_split, **kwargs)

    def hyperopt_model(self, *args):
        # overriding parent method as there's nothing to tune
        return



"""
LightGBM
-----------------
"""

from typing import Any
from functools import partial

from darts import models
import numpy as np
import numpy.typing as npt
import optuna

from oats.models._darts_simple import SimpleDartsModel


class LightGBMModel(SimpleDartsModel):
    """LightGBM Model

    Using regression via gradient boosted trees as a predictor. Anomalies scores are deviations from predictions.

    Reference: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.gradient_boosted_model.html
    """

    def __init__(
        self,
        window: int = 10,
        n_steps: int = 1,
        lags: int = 1,
        val_split: float = 0.0,
        **kwargs
    ):
        """
        initialization also accepts any parameters used by: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.gradient_boosted_model.html

        Args:
            window (int, optional): rolling window size to feed into the predictor. Defaults to 10.
            n_steps (int, optional): number of steps to predict forward. Defaults to 1.
            lags (int, optional): number of lags. Defaults to 1.
            val_split (float, optional): proportion of data points reserved for validation; only used if using auto-tuning (not tested). Defaults to 0.
        """

        model_cls = models.LightGBMModel

        super().__init__(model_cls, window, n_steps, lags, val_split, **kwargs)

    def _model_objective(self, trial, train_data: npt.NDArray[Any]):
        params = {
            "lambda_l1": trial.suggest_loguniform("lambda_l1", 1e-8, 10.0),
            "lambda_l2": trial.suggest_loguniform("lambda_l2", 1e-8, 10.0),
            "num_leaves": trial.suggest_int("num_leaves", 2, 256),
            "feature_fraction": trial.suggest_uniform("feature_fraction", 0.4, 1.0),
            "bagging_fraction": trial.suggest_uniform("bagging_fraction", 0.4, 1.0),
            "bagging_freq": trial.suggest_int("bagging_freq", 1, 7),
            "min_child_samples": trial.suggest_int("min_child_samples", 5, 100),
        }

        return self._get_hyperopt_res(params, train_data)



"""
Temporal Convolution Networks (TCN)
-----------------
"""
from typing import Any
from functools import partial

from darts import models
import numpy as np
import numpy.typing as npt
import optuna

from oats.models._darts_model import DartsModel


class TCNModel(DartsModel):
    """TCN Model (Temporal Convolution Network)

    Using TCN as a predictor. Anomalies scores are deviations from predictions.

    Reference: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.tcn_model.html
    """

    def __init__(
        self,
        window: int = 10,
        n_steps: int = 1,
        use_gpu: bool = 1,
        val_split: float = 0.2,
        **kwargs
    ):
        """
        initialization also accepts any parameters used by: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.tcn_model.html

        Args:
            window (int, optional): rolling window size to feed into the predictor. Defaults to 10.
            n_steps (int, optional): number of steps to predict forward. Defaults to 1.
            use_gpu (bool, optional): whether to use GPU. Defaults to False.
            val_split (float, optional): proportion of data points reserved for validation. Defaults to 0.2.
        """

        model = models.TCNModel

        super().__init__(model, window, n_steps, use_gpu, val_split, **kwargs)

    def _model_objective(self, trial, train_data: npt.NDArray[Any]):
        params = {
            "kernel_size": trial.suggest_int(
                "kernel_size", 2, min(32, self.window - 1)
            ),
            "num_filters": trial.suggest_int("num_filters", 2, 8),
            "weight_norm": trial.suggest_categorical("weight_norm", [True, False]),
            "dilation_base": trial.suggest_int("dilation_base", 1, 4),
            "dropout": trial.suggest_float("dropout", 0.0, 0.3),
        }

        return self._get_hyperopt_res(params, train_data)



"""
Transformer
-----------------
"""
from typing import Any
from functools import partial

from darts import models
import numpy as np
import numpy.typing as npt
import optuna


from oats.models._darts_model import DartsModel


class TransformerModel(DartsModel):
    """Transformer Model

    Using Transformer as a predictor. Anomalies scores are deviations from predictions.

    Reference: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.transformer_model.html
    """

    def __init__(
        self,
        window: int = 10,
        n_steps: int = 1,
        use_gpu: bool = False,
        val_split: float = 0.2,
    ):
        """
        initialization also accepts any parameters used by: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.transformer_model.html

        Args:
            window (int, optional): rolling window size to feed into the predictor. Defaults to 10.
            n_steps (int, optional): number of steps to predict forward. Defaults to 1.
            use_gpu (bool, optional): whether to use GPU. Defaults to False.
            val_split (float, optional): proportion of data points reserved for validation. Defaults to 0.2.
        """

        model = models.TransformerModel
        super().__init__(model, window, n_steps, use_gpu, val_split)

    def _model_objective(self, trial, train_data: npt.NDArray[Any]):
        """
        params = {
            # "nhead": trial.suggest_int("nhead", 2, 8, 2),
        }

        dep_params = {
            "dim_feedforward": trial.suggest_int("dim_feedforward", 256, 1024),
            "num_encoder_layers": trial.suggest_int("num_encoder_layers", 2, 10),
            "num_decoder_layers": trial.suggest_int("num_decoder_layers", 2, 10),
            "d_model": trial.suggest_int("d_model", 32, 256),
        }

        params.update(dep_params)
        """

        # TODO: figure out why this isn't working
        return 0



"""
Random Forest
-----------------
"""
from typing import Any
from functools import partial

from darts import models
import numpy as np
import numpy.typing as npt
import optuna


from oats.models._darts_simple import SimpleDartsModel


class RandomForestModel(SimpleDartsModel):
    """Random Forest Model

    Using random forest regression as a predictor. Anomalies scores are deviations from predictions.

    Reference: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.random_forest.html
    """

    def __init__(
        self,
        window: int = 10,
        n_steps: int = 1,
        lags: int = 1,
        val_split: float = 0.2,
        **kwargs
    ):
        """
        initialization also accepts any parameters used by: https://unit8co.github.io/darts/generated_api/darts.models.forecasting.random_forest.html

        Args:
            window (int, optional): rolling window size to feed into the predictor. Defaults to 10.
            n_steps (int, optional): number of steps to predict forward. Defaults to 1.
            lags (int, optional): number of lags. Defaults to 1.
            val_split (float, optional): proportion of data points reserved for validation; only used if using auto-tuning (not tested). Defaults to 0.
        """

        model = models.RandomForest

        super().__init__(model, window, n_steps, lags, val_split, **kwargs)

    def _model_objective(self, trial, train_data: npt.NDArray[Any]):
        params = {
            "n_estimators": trial.suggest_int("n_estimators", 30, 1000),
            "max_features": trial.suggest_categorical(
                "max_features", ["auto", "sqrt", "log2"]
            ),
            "max_depth": trial.suggest_int("max_depth", 1, 5000),
            # "ccp_alpha": trial.suggest_float("ccp_alpha", 0.0, 2e-2)
        }

        return self._get_hyperopt_res(params, train_data)



"""
Quantile
-----------------
"""


import numpy as np
from numpy.lib.stride_tricks import sliding_window_view

from oats.models._base import Model


class QuantileModel(Model):
    """Quantile Model

    Rule-based model where we label timestep as anomalous if current observation is larger than past `window` onbservations at `quantile`.
    """

    def __init__(self, window: int = 10, quantile: float = 0.98):
        """
        Args:
            window (int, optional): rolling window size to feed into the predictor. Defaults to 10.
            quantile (float, optional): threshold level for anomaly labelling. Defaults to 0.98.
        """
        self.window = window
        self.quantile = quantile
        pass

    def fit(self, *args, **kwargs):
        return

    def get_scores(self, data):
        if data.ndim > 1 and data.shape[1] > 1:
            return self._handle_multivariate(data, [self] * data.shape[1])

        E = np.zeros(len(data))

        for i in range(len(data)):
            if i > self.window - 1:
                window_threshold = np.percentile(
                    data[i - self.window : i],
                    self.quantile * 100,
                )
                E[i] = 1 if data[i] > window_threshold else 0

        return E



"""
Spectral Residual
-------------------
"""


# Copyright (c) 2019 Takahiro Yoshinaga

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy import stats

from oats.preprocessor._base import Preprocessor


def _series_filter(values, kernel_size=3):
    """
    Filter a time series. Practically, calculated mean value inside kernel size.
    As math formula, see https://docs.opencv.org/2.4/modules/imgproc/doc/filtering.html.

    Args:
        values:
        kernel_size:

    Returns:
        The list of filtered average
    """
    filter_values = np.cumsum(values, dtype=float)

    filter_values[kernel_size:] = (
        filter_values[kernel_size:] - filter_values[:-kernel_size]
    )
    filter_values[kernel_size:] = filter_values[kernel_size:] / kernel_size

    for i in range(1, kernel_size):
        filter_values[i] /= i + 1

    return filter_values


def _extrapolate_next(values):
    """
    Extrapolates the next value by sum up the slope of the last value with previous values.
    :param values: a list or numpy array of time-series
    :return: the next value of time-series
    """

    last_value = values[-1]
    slope = [(last_value - v) / (i+1e-5) for (i, v) in enumerate(values[::-1])]
    slope[0] = 0
    next_values = last_value + np.cumsum(slope)

    return next_values


def _merge_series(values, extend_num=5, forward=5):

    next_value = _extrapolate_next(values)[forward]
    extension = [next_value] * extend_num

    if isinstance(values, list):
        marge_values = values + extension
    else:
        marge_values = np.append(values, extension)
    return marge_values


class SpectralResidual(Preprocessor):
    """Generates a saliency map via spectral residual.

    Inspired from:
        Hansheng Ren, Bixiong Xu, Yujing Wang, Chao Yi, Congrui Huang, Xiaoyu Kou, Tony Xing, Mao Yang, Jie Tong, Qi Zhang.
        "Time-Series Anomaly Detection Service at Microsoft." arXiv preprint arXiv:1906.03821 (2019).

    Implementation: https://github.com/y-bar/ml-based-anomaly-detection
    """

    def __init__(self, amp_window_size=16, series_window_size=16, score_window_size=32):
        self.amp_window_size = amp_window_size
        self.series_window_size = series_window_size
        self.score_window_size = score_window_size

    def _transform_silency_map(self, values):
        """
        Transform a time-series into spectral residual, a method adopted from computer vision.
        For example, See https://github.com/uoip/SpectralResidualSaliency.

        Args:
            values: a list or numpy array of float values.

        Returns:
            Silency map and spectral residual
        """

        freq = np.fft.fft(values)
        mag = np.sqrt(freq.real**2 + freq.imag**2)
        spectral_residual = np.exp(
            np.log(mag) - _series_filter(np.log(mag), self.amp_window_size)
        )

        freq.real = freq.real * spectral_residual / mag
        freq.imag = freq.imag * spectral_residual / mag

        silency_map = np.fft.ifft(freq)
        return silency_map

    def _transform_spectral_residual(self, values):
        silency_map = self._transform_silency_map(values)
        spectral_residual = np.sqrt(silency_map.real**2 + silency_map.imag**2)
        return spectral_residual

    def transform(self, values: ArrayLike, type: str = "avg") -> NDArray:
        """
        Transform series using Spectral Residual

        Args:
            values: timeseries
            type: filter type in `["avg", "abs", "chisq"]`

        Returns:

        """
        multivar = True if values.ndim > 1 and values.shape[1] > 1 else False
        if multivar:
            return self._handle_multivariate(values, [self] * values.shape[1])

        extended_series = _merge_series(
            values, self.series_window_size, self.series_window_size
        )
        mag = self._transform_spectral_residual(extended_series)[: len(values)]

        if type == "avg":
            ave_filter = _series_filter(mag, self.score_window_size)
            score = (mag - ave_filter) / ave_filter
        elif type == "abs":
            ave_filter = _series_filter(mag, self.score_window_size)
            score = np.abs(mag - ave_filter) / ave_filter
        elif type == "chisq":
            score = stats.chi2.cdf((mag - np.mean(mag)) ** 2 / np.var(mag), df=1)
        else:
            raise ValueError("No type!")
        return score

    def fit(self, *arg, **kwargs) -> None:
        return



"""
Thresholding
-----------------
"""

from oats.threshold.pot import POTThreshold
from oats.threshold.spot import SPOTThreshold
from oats.threshold.quantile import QuantileThreshold
from oats.threshold.jenks import JenksThreshold



from oats._utils import graph_data


class Visualizer:
    def __init__(self, series, labels, train_len):
        self.p = graph_data(series, labels, train_len)
        self.ax1, self.ax2 = self.p.get_axes()

        self.idx = 0

    def add_result(self, anoms: list, model_name: str):
        xmin = 0.125
        ymin = 0.1 - (1 + self.idx) * 0.05
        dx = 0.775
        dy = 0.05

        ax_new = p.add_axes([xmin, ymin, dx, dy], anchor="SW", sharex=self.ax1)
        ax_new.text(10, 0.5, model_name, fontfamily="monospace", va="center")

        ax_new.set_xticklabels([])
        ax_new.set_xticks([])
        ax_new.set_yticklabels([])
        ax_new.set_yticks([])

        ax_new.vlines(anoms, 0, 1, colors="red")
        self.idx += 1



"""
Qualitative Metrics
-----------------
"""

import numpy as np
import scipy.signal as signal

from oats.scorer._base import Scorer


class QualitativeMetrics(Scorer):
    """Unsupervised qualitative metrics used to access the quality of the anomaly detection algorithm"""

    def __init__(self, window=10):
        """
        Args:
            window (int, optional): Window sized used to compute `diff_mean_trend` and `diff_mid_avg`. Defaults to 10.
        """
        self.data = np.array([])
        self.preds = np.array([])

        self.window = window

    def process(self, data, preds):
        if data.ndim > 1 and data.shape[1] == 1:
            data = data.flatten()

        if len(data) != len(preds):
            raise ValueError(
                f"Data of length {len(data)} does not match preds of length {len(preds)}"
            )

        existing_n_feat = self.data.shape[1] if self.data.ndim > 1 else 1
        incoming_n_feat = data.shape[1] if data.ndim > 1 else 1
        if len(self.data) > 0 and existing_n_feat != incoming_n_feat:
            raise ValueError(
                f"Unable to process incoming data of shape {data.shape} with existing data of shape {self.data.shape}"
            )

        if data.ndim > 1 and data.shape[1] > 1:
            self.data = np.vstack([self.data] * data.shape[1]).T

        self.data = np.append(self.data, data, axis=0)
        self.preds = np.append(self.preds, preds)

    @property
    def num_anom(self):
        """Number of predicted anomalies, should be low (as anomalies are rare)"""
        return self.preds.sum()

    @property
    def pct_anom(self):
        """Percentage of predicted anomalies, should be low (as anomalies are rare)"""
        return self.num_anom / len(self.preds)

    @property
    def _pred_anomalies(self):
        return self.data[self.preds == 1]

    @property
    def _pred_non_anomalies(self):
        return self.data[self.preds == 0]

    @property
    def avg_anom_dist_from_mean(self):
        """Distance of predicted anomalies to the mean of original data, should be high; useful for series with a lot of global point anomalies"""
        return np.abs(self._pred_anomalies - self.data.mean(axis=0)).mean()

    @property
    def avg_cycles_delta_between_anom(self):
        """Average time between anomalies, should be high, as anomalies should be occuring far apart (for point anomalies)"""
        if self.num_anom in (0, 1, len(self.data)):
            return 0
        return np.diff(np.where(self.preds == 1)[0]).mean()

    @property
    def max_range_non_anom(self):
        """The tightness of data from predicted non-anomalies, similar to the idea of `avg_anom_dist_from_mean`; should be low"""
        if self.num_anom in (0, len(self.data)):
            return 1e5
        return (
            np.abs(self._pred_non_anomalies.max() - self._pred_non_anomalies.min())
        ).mean()

    def _get_mid_avg_filter(self):
        # make sure window is odd
        if self.window % 2 == 0:
            self.window += 1
        padding = self.window // 2

        # local difference filter
        fil = self.window
        fil = np.full((self.window), -1 / (self.window - 1))
        fil[padding] = 1
        # if self.data.ndim > 1:
        # fil = np.tile(fil, (self.data.shape[1], 1)).T

        return fil

    @property
    def diff_mean_trend(self):
        """The trend (gradient) of predicted anomalies vs the trend of surrounding points, should be high"""
        if not 0 < self.num_anom < len(self.preds):
            return 0

        fil = self._get_mid_avg_filter()
        padding = self.window // 2

        # abs-trend avg
        diffs = []
        data = self.data if self.data.ndim > 1 else self.data[:, np.newaxis]
        for arr in data.T:
            grads = signal.savgol_filter(arr, self.window, 1, deriv=1, axis=0)
            grads = np.abs(grads)
            conv = np.abs(signal.convolve(grads, fil, mode="valid"))
            conv = np.pad(conv, (padding, padding), mode="edge")
            diffs.append(
                (
                    conv[self.preds == 1].mean(axis=0)
                    - conv[self.preds == 0].mean(axis=0)
                ).sum()
            )

        return np.mean(diffs)

    @property
    def diff_mid_avg(self):
        """The value of predicted anomalies vs the average of surrounding values, should be high"""
        if not 0 < self.num_anom < len(self.preds):
            return 0

        fil = self._get_mid_avg_filter()
        padding = self.window // 2

        # ~= midpoint minus avg. of sides
        diffs = []
        data = self.data if self.data.ndim > 1 else self.data[:, np.newaxis]
        for arr in data.T:
            conv = np.abs(signal.convolve(arr, fil, mode="valid"))
            conv = np.pad(conv, (padding, padding), mode="edge")
            diffs.append(
                conv[self.preds == 1].mean(axis=0)
                - conv[self.preds == 0].mean(axis=0).sum()
            )

        return np.mean(diffs)



"""
Streaming Peaks-Over-Threshold (SPOT)
-----------------
"""
import time
import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
from statsmodels.robust.scale import huber
from scipy.stats import genpareto
from scipy.stats import norm

from oats.threshold._base import Threshold


class SPOTThreshold(Threshold):
    """Streaming version of POT
    Fit the tails of the data with Generalized Pareto Distribution (GPD).
    Find the threshold where `P(thres) < q`.
    Usual values for q is 1e-3 to 1e-6.

    `fit()` is required to compute the initial threshold for the first `n` points.

    Siffer, Alban and Fouque, Pierre-Alain and Termier, Alexandre and Largouet, Christine
    "Anomaly Detection in Streams with Extreme Value Theory"
    https://doi.org/10.1145/3097983.3098144
    """

    def __init__(
        self,
        q: float = 1e-4,
        level: float = 0.95,
        memory: int = 2000,
        support: float = 0,
        init_cutoff: float = 1,
        robust: bool = False,
        estimator: str = "MoM",
        **kwargs
    ):
        """
        Args:
            q (float, optional): q level such that `P(threshold) < q`. Defaults to 1e-4.
            level (float, optional): threshold to fit tail distribution. Defaults to 0.95.
            memory (int, optional): how many data points for the tails. Defaults to 2000.
            support (float, optional): used for method of moments estimation to prevent a bad fit (MoM is fast but unreliable); threshold is computed as the maximum of `GPD` and `support * HalfNormal`. Defaults to 0.
            init_cutoff (float, optional): used for method of moments estimation to prevent a bad fit (MoM is not robust against outliers); only bottom `init_cutoff` deciles are used to fit the GPD distribution. Defaults to 1.
            robust (bool, optional): (experimental) using robust estimates of `mean` and `variance` to fit GPD using MoM. Defaults to False.
            estimator (str, optional): estimator used to fit GPD distribution; accepts 'MoM' or 'MLE'. Defaults to "MoM".
        """
        self.support = support
        self.init_cutoff = init_cutoff
        self.robust = robust

        self.q = q
        self.level = level
        self.memory = memory

        # Rolling Data
        self.X = np.array([])

        # Peak Set
        self.Y = np.array([])

        # len(data set)
        self.n = 0

        # I -> indices of peaks in order to maintain the correct n_peak/n ratio
        self.I = np.array([])

        # thresholds
        self.percentile_thres = 0
        self.spot_thres = 0

        self.estimator = estimator

        self._thresholders = None

    @property
    def adjusted_n(self):
        return self.n - self.I[0]

    # TODO: typing
    #       data -> np.NDArray
    def fit(self, data):
        multivar = True if data.ndim > 1 and data.shape[1] > 1 else False
        if multivar:
            self._thresholders = self._pseudo_mv_fit(data)
            return

        for x in data:
            self._add_mem(x)

        self.spot_thres = self._spot_initialize(data)

    def _step(self, x_t) -> float:
        """
        Stepping + Updating SPOT
        """
        self._add_mem(x_t)
        new_p_thres = _GPDThreshold.get_tail_threshold(self.X, self.level)
        self.spot_thres += new_p_thres - self.percentile_thres
        self.percentile_thres = new_p_thres

        if self.X[-1] > self.spot_thres:
            """erasing anomalous data point from history"""
            self.X[-1] = 0

        elif self.X[-1] > self.percentile_thres:
            # add obs to peak set
            y_i = self.X[-1] - self.percentile_thres
            self._add_peak(y_i)

            sigma, gamma = _GPDThreshold.get_gpd_params(
                self.Y, robust=self.robust, estimator=self.estimator
            )
            self.spot_thres = max(
                _GPDThreshold.calc_spot_threshold(
                    self.percentile_thres,
                    sigma,
                    gamma,
                    self.adjusted_n,
                    self.Y.size,
                    self.q,
                ),
                _GPDThreshold.calc_half_normal_threshold(
                    self.percentile_thres,
                    self.Y.std(ddof=1),
                    self.q,
                    support=self.support,
                ),
            )

        return self.spot_thres

    def get_threshold(self, data):
        multivar = True if data.ndim > 1 and data.shape[1] > 1 else False
        if multivar:
            if not self._thresholders:
                self._thresholders = [self] * data.shape[1]
            return self._handle_multivariate(data, self._thresholders)

        ret = np.array([])
        for x in data:
            ret = np.append(ret, self._step(x))

        return ret

    @property
    def n_peaks(self):
        return self.Y.size

    def _spot_initialize(self, S):
        self.percentile_thres = _GPDThreshold.get_tail_threshold(S, self.level)

        Y = _GPDThreshold.get_peaks(S, self.percentile_thres)

        if self.init_cutoff < 1:
            cutoff = _GPDThreshold.get_tail_threshold(Y, self.init_cutoff)
            Y = Y[Y < cutoff]

        for y in Y:
            self._add_peak(y)

        sigma, gamma = _GPDThreshold.get_gpd_params(
            Y, robust=self.robust, estimator=self.estimator
        )

        n_y = len(Y)
        n = len(S)

        spot_thres = max(
            _GPDThreshold.calc_spot_threshold(
                self.percentile_thres, sigma, gamma, n, n_y, self.q
            ),
            _GPDThreshold.calc_half_normal_threshold(
                self.percentile_thres, Y.std(ddof=1), self.q, support=self.support
            ),
        )

        return spot_thres

    def _add_peak(self, y):
        if self.Y.size <= self.memory:
            self.Y = np.append(self.Y, y)
        else:
            self.Y = np.roll(self.Y, -1)
            self.Y[-1] = y

        if self.I.size <= self.memory + 1:
            self.I = np.append(self.I, self.n)
        else:
            self.I = np.roll(self.I, -1)
            self.I[-1] = self.n

    def _add_mem(self, x):
        if self.X.size <= self.memory:
            self.X = np.append(self.X, x)
        else:
            self.X = np.roll(self.X, -1)
            self.X[-1] = x
        self.n += 1


# SPOT based on MoM-Estimation Helper Class
class _GPDThreshold:
    @classmethod
    def get_tail_threshold(cls, data, level) -> float:
        return np.quantile(data, level)

    @classmethod
    def get_peaks(cls, data, threshold, n_sample_min=5):
        x = data.copy()
        x = x[x >= threshold]

        if len(x) < n_sample_min:
            threshold *= 0.95
            threshold = 0 if threshold < 1e-3 else threshold
            return cls.get_peaks(data, threshold, n_sample_min)

        return x - threshold

    @classmethod
    def get_gpd_params_mv(cls, mean, var):
        if estimator == "MoM":
            if var == 0:
                return 1, 1

            sigma = mean / 2 * (1 + mean**2 / var)
            gamma = 1 / 2 * (1 - (mean**2 / var))

        return sigma, gamma

    @classmethod
    def get_gpd_params(cls, peaks, robust=False, estimator="MoM"):
        y = peaks.copy()
        if estimator != "MoM":
            mu, sigma, gamma = genpareto.fit(y)
            return sigma, gamma

        if robust:
            try:
                huber.maxiter = 1000
                huber.tol = 5e-2
                mu, std = huber(y)
                var_y = std**2
            except ValueError:
                mu, var_y = y.mean(), y.var(ddof=1)
        else:
            mu, var_y = y.mean(), y.var(ddof=1)

        sigma = mu / 2 * (1 + mu**2 / var_y)
        gamma = 1 / 2 * (1 - (mu**2 / var_y))

        return sigma, gamma

    @classmethod
    def calc_spot_threshold(cls, initial_threshold, sigma, gamma, n, n_peaks, q):
        delta = sigma / gamma * ((q * n / n_peaks) ** (-gamma) - 1)
        new_thres = initial_threshold + delta

        return new_thres

    @classmethod
    def calc_half_normal_threshold(cls, initial_threshold, std_dev, q, support=1):
        return initial_threshold + std_dev * norm.ppf(((1 - q) + 1) / 2) * support



"""
Isolation Forest
-----------------
"""

from pyod.models.iforest import IForest
from oats.models._pyod_model import PyODModel


class IsolationForestModel(PyODModel):
    def __init__(self, window=10, **kwargs):

        model_cls = IForest

        super().__init__(model_cls, window, **kwargs)

    def hyperopt_model(self, *args):
        # overriding parent method as there's nothing to tune
        return



"""
TranAD
-----------------
"""

from typing import Any

import numpy as np
import numpy.typing as npt
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from importlib_metadata import version
from oats.models._base import Model
from torch.nn import TransformerDecoder, TransformerEncoder
from torch.utils.data import DataLoader, Dataset, TensorDataset
from oats._utils.dlutils import *
from darts.timeseries import TimeSeries
from darts.dataprocessing.transformers import Scaler


# Proposed Model + Self Conditioning + Adversarial + MAML (TKDE 21)
class _TranAD(nn.Module):
    def __init__(self, feats, window=100):
        super(_TranAD, self).__init__()
        self.name = "TranAD"
        self.lr = 0.005
        self.batch = 128
        self.n_feats = feats
        self.n_window = window
        self.n = self.n_feats * self.n_window
        self.pos_encoder = PositionalEncoding(2 * feats, 0.1, self.n_window)
        encoder_layers = TransformerEncoderLayer(
            d_model=2 * feats, nhead=feats, dim_feedforward=16, dropout=0.1
        )
        self.transformer_encoder = TransformerEncoder(encoder_layers, 1)
        decoder_layers1 = TransformerDecoderLayer(
            d_model=2 * feats, nhead=feats, dim_feedforward=16, dropout=0.1
        )
        self.transformer_decoder1 = TransformerDecoder(decoder_layers1, 1)
        decoder_layers2 = TransformerDecoderLayer(
            d_model=2 * feats, nhead=feats, dim_feedforward=16, dropout=0.1
        )
        self.transformer_decoder2 = TransformerDecoder(decoder_layers2, 1)
        self.fcn = nn.Sequential(nn.Linear(2 * feats, feats), nn.Sigmoid())

    def encode(self, src, c, tgt):
        src = torch.cat((src, c), dim=2)
        src = src * math.sqrt(self.n_feats)
        src = self.pos_encoder(src)
        memory = self.transformer_encoder(src)
        tgt = tgt.repeat(1, 1, 2)
        return tgt, memory

    def forward(self, src, tgt):
        # Phase 1 - Without anomaly scores
        c = torch.zeros_like(src)
        x1 = self.fcn(self.transformer_decoder1(*self.encode(src, c, tgt)))
        # Phase 2 - With anomaly scores
        c = (x1 - src) ** 2
        x2 = self.fcn(self.transformer_decoder2(*self.encode(src, c, tgt)))
        return x1, x2


class TranADModel(Model):
    """TranAD Model

    Tuli, Shreshth and Casale, Giuliano and Jennings, Nicholas R
    "TranAD: Deep Transformer Networks for Anomaly Detection in Multivariate Time Series Data"

    Implementation from reference: https://github.com/imperial-qore/TranAD
    """

    def __init__(
        self,
        window: int = 100,
        use_gpu: bool = False,
        val_split: float = 0.2,
    ):
        """
        Args:
            window (int, optional): rolling window size to feed into the predictor. Defaults to 10.
            use_gpu (bool, optional): whether to use GPU. Defaults to False.
            val_split (float, optional): proportion of data points reserved for validation. Defaults to 0.2.
        """
        # initiate parameters
        self.model_cls = _TranAD
        self.window = window
        self.use_gpu = use_gpu
        self.val_split = val_split

    @property
    def _model_name(self):
        return type(self).__name__

    def _init_model(self, data_dims: int):
        self.model = self.model_cls(data_dims).double()
        self.optimizer = torch.optim.AdamW(
            self.model.parameters(), lr=self.model.lr, weight_decay=1e-5
        )
        self.scheduler = torch.optim.lr_scheduler.StepLR(self.optimizer, 5, 0.9)

        return

    def fit(self, train_data, epochs=5):
        self.epochs = epochs
        train_data = train_data if train_data.ndim > 1 else train_data[:, np.newaxis]
        self._init_model(train_data.shape[-1])

        train_data = torch.tensor(train_data)
        trainD = self._convert_to_windows(train_data)

        for e in range(epochs):
            self._backprop(trainD)  # returns loss, lr

        return

    def get_scores(self, test_data):
        test_data = test_data if test_data.ndim > 1 else test_data[:, np.newaxis]
        test_data = torch.tensor(test_data)
        feats = test_data.shape[-1]
        data_len = len(test_data)
        test_data = self._convert_to_windows(test_data)
        dataloader = self._get_dataloader(test_data, data_len)

        l = nn.MSELoss(reduction="none")

        for d, _ in dataloader:
            window = d.permute(1, 0, 2)
            elem = window[-1, :, :].view(1, data_len, feats)
            z = self.model(window, elem)
            if isinstance(z, tuple):
                z = z[1]
            loss = l(z, elem)[0]

        scores = loss.detach().numpy()

        if scores.shape[1] == 1:
            scores = scores.flatten()
        return scores

    def _convert_to_windows(self, data):
        windows = []
        for i, g in enumerate(data):
            if i >= self.window:
                w = data[i - self.window : i]
            else:
                w = torch.cat([data[0].repeat(self.window - i, 1), data[0:i]])
            windows.append(w)
        windows = torch.stack(windows)
        return windows

    def _get_dataloader(self, data, bs):
        data_x = torch.DoubleTensor(data)
        dataset = TensorDataset(data_x, data_x)

        return DataLoader(dataset, batch_size=bs)

    def _backprop(self, data):
        dataloader = self._get_dataloader(data, self.model.batch)

        feats = data.shape[-1]
        l = nn.MSELoss(reduction="none")
        n = self.epochs + 1

        l1s, l2s = [], []

        for d, _ in dataloader:
            local_bs = d.shape[0]
            window = d.permute(1, 0, 2)
            elem = window[-1, :, :].view(1, local_bs, feats)
            z = self.model(window, elem)
            l1 = (
                l(z, elem)
                if not isinstance(z, tuple)
                else (1 / n) * l(z[0], elem) + (1 - 1 / n) * l(z[1], elem)
            )
            if isinstance(z, tuple):
                z = z[1]

            l1s.append(torch.mean(l1).item())
            loss = torch.mean(l1)
            self.optimizer.zero_grad()
            loss.backward(retain_graph=True)
            self.optimizer.step()

        self.scheduler.step()

        return np.mean(l1s), self.optimizer.param_groups[0]["lr"]

    def _scale_series(self, series: npt.NDArray[Any]):
        series = TimeSeries.from_values(series)

        if self.transformer is None:
            self.transformer = Scaler()
            self.transformer.fit(series)

        series = self.transformer.transform(series)

        return series.pd_dataframe().to_numpy().astype(np.float32)
